#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pytest
from schema_agents.role import Role
from schema_agents.schema import Message
import numpy as np

    
@pytest.mark.asyncio
async def test_logprobs():
    async def respond_to_user(query: str, role: Role) -> str:
        """Respond to user."""
        messages = [{"role": "user", "content": "How are you today? Anwser YES or NO, nothing else!"}]
        text, logprops = await role.llm.acompletion(messages, logprobs=True, top_logprobs=3)
        logprobs_bytes = []
        for prob in logprops:
            logprobs_bytes.extend(prob.bytes)  # Use extend instead of +=
        text_response = bytes(logprobs_bytes).decode("utf-8")
        assert text == text_response
        assert logprops[0].token == 'YES'
        assert np.round(np.exp(logprops[0].logprob),2) > 0.5

    alice = Role(
        name="Alice",
        profile="Cooker",
        goal="Your goal is to listen to user's request and propose recipes for making the most delicious meal for thanksgiving.",
        constraints=None,
        actions=[respond_to_user],
        model="gpt-4-1106-preview",
    )
    event_bus = alice.get_event_bus()
    event_bus.register_default_events()
    await alice.handle(Message(content="make something to surprise our guest from Stockholm.", role="User"))
    


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Time    : 2023/5/5 23:08
@Author  : alexanderwu
@File    : openai.py
"""
import asyncio
import time
import random
import string
from functools import wraps
from typing import NamedTuple, Union, List, Dict, Any

import httpx
import openai
from openai import AsyncOpenAI, OpenAI

from schema_agents.config import CONFIG
from schema_agents.logs import logger
from schema_agents.provider.base_gpt_api import BaseGPTAPI
from schema_agents.utils.singleton import Singleton
from schema_agents.utils.token_counter import (
    TOKEN_COSTS,
    count_message_tokens,
    count_string_tokens,
    get_max_completion_tokens,
)
from schema_agents.utils.common import EventBus, current_session
from schema_agents.schema import StreamEvent
from contextvars import copy_context

def retry(max_retries):
    def decorator(f):
        @wraps(f)
        async def wrapper(*args, **kwargs):
            for i in range(max_retries):
                try:
                    return await f(*args, **kwargs)
                except Exception:
                    if i == max_retries - 1:
                        raise
                    await asyncio.sleep(2 ** i)
        return wrapper
    return decorator


class RateLimiter:
    """Rate control class, each call goes through wait_if_needed, sleep if rate control is needed"""
    def __init__(self, rpm):
        self.last_call_time = 0
        self.interval = 1.1 * 60 / rpm  # Here 1.1 is used because even if the calls are made strictly according to time, they will still be QOS'd; consider switching to simple error retry later
        self.rpm = rpm

    def split_batches(self, batch):
        return [batch[i:i + self.rpm] for i in range(0, len(batch), self.rpm)]

    async def wait_if_needed(self, num_requests):
        current_time = time.time()
        elapsed_time = current_time - self.last_call_time

        if elapsed_time < self.interval * num_requests:
            remaining_time = self.interval * num_requests - elapsed_time
            logger.info(f"sleep {remaining_time}")
            await asyncio.sleep(remaining_time)

        self.last_call_time = time.time()


class Costs(NamedTuple):
    total_prompt_tokens: int
    total_completion_tokens: int
    total_cost: float
    total_budget: float


class CostManager(metaclass=Singleton):
    """计算使用接口的开销"""
    def __init__(self):
        self.total_prompt_tokens = 0
        self.total_completion_tokens = 0
        self.total_cost = 0
        self.total_budget = 0

    def update_cost(self, prompt_tokens, completion_tokens, model):
        """
        Update the total cost, prompt tokens, and completion tokens.

        Args:
        prompt_tokens (int): The number of tokens used in the prompt.
        completion_tokens (int): The number of tokens used in the completion.
        model (str): The model used for the API call.
        """
        self.total_prompt_tokens += prompt_tokens
        self.total_completion_tokens += completion_tokens
        cost = (
            prompt_tokens * TOKEN_COSTS[model]["prompt"]
            + completion_tokens * TOKEN_COSTS[model]["completion"]
        ) / 1000
        self.total_cost += cost
        logger.info(f"Total running cost: ${self.total_cost:.3f} | Max budget: ${CONFIG.max_budget:.3f} | "
                    f"Current cost: ${cost:.3f}, {prompt_tokens=}, {completion_tokens=}")
        CONFIG.total_cost = self.total_cost

    def get_total_prompt_tokens(self):
        """
        Get the total number of prompt tokens.

        Returns:
        int: The total number of prompt tokens.
        """
        return self.total_prompt_tokens

    def get_total_completion_tokens(self):
        """
        Get the total number of completion tokens.

        Returns:
        int: The total number of completion tokens.
        """
        return self.total_completion_tokens

    def get_total_cost(self):
        """
        Get the total cost of API calls.

        Returns:
        float: The total cost of API calls.
        """
        return self.total_cost

    def get_costs(self) -> Costs:
        """获得所有开销"""
        return Costs(self.total_prompt_tokens, self.total_completion_tokens, self.total_cost, self.total_budget)


class OpenAIGPTAPI(BaseGPTAPI, RateLimiter):
    """
    Check https://platform.openai.com/examples for examples
    """
    def __init__(self, model=None, seed=None, temperature=None, stop=None, logprobs=None, top_logprobs=None, timeout=None):
        self.__init_openai(CONFIG)
        self.model = model or CONFIG.openai_api_model
        self.temperature = float(temperature or CONFIG.openai_temperature)
        self.stop = stop or CONFIG.openai_stop
        self.logprobs = logprobs or CONFIG.openai_logprobs
        self.top_logprobs = top_logprobs or CONFIG.openai_top_logprobs
        self.timeout = timeout or CONFIG.openai_timeout
        self.seed = int(seed or CONFIG.openai_seed)
        self.auto_max_tokens = False
        self._cost_manager = CostManager()
        RateLimiter.__init__(self, rpm=self.rpm)
        logger.info(f"OpenAI API model: {self.model}")

    def __init_openai(self, config):
        if config.openai_proxy:
            openai.http_client = httpx.Client(
                proxies=config.openai_proxy,
            )
        if config.openai_api_type:
            openai.api_type = config.openai_api_type
            openai.api_version = config.openai_api_version
        self.aclient = AsyncOpenAI(
            api_key=config.openai_api_key,
            base_url=config.openai_api_base,
        )
        self.client = OpenAI(
            api_key=config.openai_api_key,
            base_url=config.openai_api_base,
        )
        self.rpm = int(config.get("RPM", 10))

    async def _achat_completion_stream(self, messages: list[dict], event_bus: EventBus=None, **kwargs) -> str:
        response = await self.aclient.chat.completions.create(
            **self._cons_kwargs(messages, functions=kwargs.get("functions")),
            stream=True,
            **kwargs
        )

        # create variables to collect the stream of chunks
        collected_chunks = []
        collected_messages = []
        collected_logprobs = []
        function_call_detected = False
        tool_call_detected = False
        func_call = {}
        tool_calls = {}
        query_id = ''.join(random.choices(string.ascii_lowercase + string.digits, k=8))
        session = current_session.get() if current_session in copy_context() else None
        acc_message = ""
        # iterate through the stream of events
        async for raw_chunk in response:
            collected_chunks.append(raw_chunk)  # save the event response
            choice0 = raw_chunk.choices[0]
            chunk_message = choice0.delta.dict()  # extract the message
            collected_messages.append(chunk_message)  # save the message
            if kwargs.get("logprobs") and choice0.logprobs and choice0.logprobs.content:
                collected_logprobs.extend(choice0.logprobs.content)
            if choice0.finish_reason == "length":
                raise RuntimeError("Incomplete model output due to max_tokens parameter or token limit")
            elif choice0.finish_reason == "content_filter":
                raise RuntimeError("Omitted content due to a flag from our content filters")

            if "tool_calls" in chunk_message and chunk_message["tool_calls"]:
                _tool_calls = chunk_message["tool_calls"]
                for _tool_call in _tool_calls:
                    if _tool_call.get("type") == "function":
                        func_call = _tool_call.get("function")
                        if event_bus:
                            event_bus.emit("stream", StreamEvent(type="function_call", query_id=query_id, session=session, name=func_call["name"], arguments=func_call.get("arguments", ""), status="start"))
                    if _tool_call['index'] not in tool_calls:
                        tool_calls[_tool_call['index']] = _tool_call
                    else:
                        func_call = tool_calls[_tool_call['index']]['function']
                        func_call['arguments'] += _tool_call['function']['arguments']
                        if event_bus:
                            event_bus.emit("stream", StreamEvent(type="function_call", query_id=query_id, session=session, name=func_call["name"], arguments=_tool_call['function']['arguments'], status="in_progress"))
                tool_call_detected = True
                
            elif "function_call" in chunk_message and chunk_message["function_call"]:
                if "name" in chunk_message["function_call"] and chunk_message["function_call"]["name"]:
                    func_call["name"] = chunk_message["function_call"]["name"]
                    if event_bus:
                        event_bus.emit("stream", StreamEvent(type="function_call", query_id=query_id, session=session, name=func_call["name"], arguments=func_call.get("arguments", ""), status="start"))
                if "arguments" in chunk_message["function_call"]:
                    if "arguments" not in func_call:
                        func_call["arguments"] = ""
                    func_call["arguments"] += chunk_message["function_call"]["arguments"]
                    if event_bus:
                        event_bus.emit("stream", StreamEvent(type="function_call", query_id=query_id, session=session, name=func_call["name"], arguments=chunk_message["function_call"]["arguments"], status="in_progress"))
                function_call_detected = True
            elif "content" in chunk_message and chunk_message["content"]:
                if event_bus:
                    acc_message += chunk_message["content"]
                    if acc_message == chunk_message["content"]:
                        event_bus.emit("stream", StreamEvent(type="text", query_id=query_id, session=session, content=chunk_message["content"], status="start"))
                    else:
                        event_bus.emit("stream", StreamEvent(type="text", query_id=query_id, session=session, content=chunk_message["content"], status="in_progress"))

            if event_bus:
                if function_call_detected:
                    if choice0.finish_reason in ["function_call", "stop"]:
                        event_bus.emit("function_call", func_call)
                        event_bus.emit("stream", StreamEvent(type="function_call", query_id=query_id, session=session, name=func_call["name"], arguments=func_call["arguments"], status="finished"))
                elif tool_call_detected:
                    if choice0.finish_reason in ["tool_calls", "stop"]:
                        for _tool_call in tool_calls.values():
                            if _tool_call.get("type") == "function":
                                func_call = _tool_call.get("function")
                                event_bus.emit("function_call", func_call)
                                event_bus.emit("stream", StreamEvent(type="function_call", query_id=query_id, session=session, name=func_call["name"], arguments=func_call.get("arguments", ""), status="finished"))
        if function_call_detected:
            full_reply_content = full_reply_content = {"type": "function_call", "function_call": func_call}
            if raw_chunk.system_fingerprint:
                full_reply_content['system_fingerprint'] = raw_chunk.system_fingerprint
            usage = self._calc_usage(messages, f"{func_call['name']}({func_call['arguments']})", functions=kwargs.get("functions", None))
        elif tool_call_detected:
            full_reply_content = {"type": "tool_calls", "tool_calls": [tool_calls[k] for k in sorted(list(tool_calls.keys()))]}
            if raw_chunk.system_fingerprint:
                full_reply_content['system_fingerprint'] = raw_chunk.system_fingerprint
            _msg = ""
            for _tool_call in tool_calls.values():
                func_call = _tool_call['function']
                _msg += f"{func_call['name']}({func_call['arguments']})\n"
            tools = kwargs["tools"]
            functions = [tool['function'] for tool in tools]
            usage = self._calc_usage(messages, _msg, functions=functions)
        else:
            full_reply_content = ''.join([m.get('content', '') for m in collected_messages if m.get('content')])
            usage = self._calc_usage(messages, full_reply_content, functions=kwargs.get("functions", None))
            if event_bus:
                event_bus.emit("stream", StreamEvent(type="text", query_id=query_id, session=session, content=full_reply_content, status="finished"))
        self._update_costs(usage)
        if event_bus:
            event_bus.emit("completion", full_reply_content)
        if collected_logprobs:
            return full_reply_content, collected_logprobs
        return full_reply_content


    def _cons_kwargs(self, messages: list[dict], functions: list[dict]=None) -> dict:
        kwargs = {
            "messages": messages,
            "max_tokens": self.get_max_tokens(messages, functions=functions),
            "n": 1,
            "temperature": self.temperature,
            "timeout": self.timeout,
            "seed": self.seed,
        }
        if self.stop:
            kwargs["stop"] = self.stop
        if self.logprobs:
            kwargs["logprobs"] = self.logprobs
        if self.top_logprobs:
            kwargs["top_logprobs"] = self.top_logprobs

        if CONFIG.openai_api_type == "azure":
            if CONFIG.deployment_name and CONFIG.deployment_id:
                raise ValueError("You can only use one of the `deployment_id` or `deployment_name` model")
            elif not CONFIG.deployment_name and not CONFIG.deployment_id:
                raise ValueError("You must specify `DEPLOYMENT_NAME` or `DEPLOYMENT_ID` parameter")
            kwargs_mode = (
                {"engine": CONFIG.deployment_name}
                if CONFIG.deployment_name
                else {"deployment_id": CONFIG.deployment_id}
            )
        else:
            kwargs_mode = {"model": self.model}
        kwargs.update(kwargs_mode)
        return kwargs

    async def _achat_completion(self, messages: list[dict], event_bus: EventBus=None, **kwargs) -> dict:
        kwargs.update(self._cons_kwargs(messages, functions=kwargs.get("functions")))
        rsp = await self.aclient.chat.completions.create(**kwargs)
        self._update_costs(rsp.usage.dict())
        if event_bus:
            event_bus.emit("completion", rsp)
        return rsp

    def _chat_completion(self, messages: list[dict], event_bus: EventBus=None) -> dict:
        rsp = self.client.chat.completions.create(**self._cons_kwargs(messages))
        self._update_costs(rsp)
        if event_bus:
            event_bus.emit("completion", rsp)
        return rsp

    def completion(self, messages: list[dict], event_bus: EventBus=None) -> dict:
        rsp = self._chat_completion(messages, event_bus=event_bus)
        return rsp

    async def acompletion(self, messages: list[dict], event_bus: EventBus=None, **kwargs) -> dict:
        rsp = await self._achat_completion_stream(messages, event_bus=event_bus, **kwargs)
        return rsp
    
    async def acompletion_function(self, messages: list[dict], functions: List[Dict[str, Any]]=None, function_call: Union[str, Dict[str, str]]=None, event_bus: EventBus=None, **kwargs) -> dict:
        rsp = await self._achat_completion_stream(messages, functions=functions, function_call=function_call, event_bus=event_bus, **kwargs)
        return rsp

    async def acompletion_tool(self, messages: list[dict], tools: List[Dict[str, Any]]=None, tool_choice: Union[str, Dict[str, str]]=None, event_bus: EventBus=None, **kwargs) -> dict:
        rsp = await self._achat_completion_stream(messages, tools=tools, tool_choice=tool_choice, event_bus=event_bus, **kwargs)
        return rsp

    async def acompletion_text(self, messages: list[dict], stream=False, event_bus: EventBus=None, **kwargs) -> str:
        """when streaming, print each token in place."""    
        if stream:
            return await self._achat_completion_stream(messages, event_bus=event_bus, **kwargs)
        rsp = await self._achat_completion(messages, event_bus=event_bus, **kwargs)
        if "logprobs" in kwargs:
            return self.get_choice_text(rsp), self.get_choice_logprobs(rsp)

        return self.get_choice_text(rsp)

    def _calc_usage(self, messages: list[dict], rsp: str, functions: List[dict]=None) -> dict:
        usage = {}
        prompt_tokens = count_message_tokens(messages, self.model, functions=functions)
        completion_tokens = count_string_tokens(rsp, self.model)
        usage['prompt_tokens'] = prompt_tokens
        usage['completion_tokens'] = completion_tokens
        return usage

    async def acompletion_batch(self, batch: list[list[dict]], event_bus: EventBus=None) -> list[dict]:
        split_batches = self.split_batches(batch)
        all_results = []

        for small_batch in split_batches:
            logger.info(small_batch)
            await self.wait_if_needed(len(small_batch))

            future = [self.acompletion(prompt, event_bus=event_bus) for prompt in small_batch]
            results = await asyncio.gather(*future)
            logger.info(results)
            all_results.extend(results)

        return all_results

    async def acompletion_batch_text(self, batch: list[list[dict]], event_bus: EventBus=None) -> list[str]:
        raw_results = await self.acompletion_batch(batch, event_bus=event_bus)
        results = []
        for idx, raw_result in enumerate(raw_results, start=1):
            result = self.get_choice_text(raw_result)
            results.append(result)
            logger.info(f"Result of task {idx}: {result}")
        return results

    def _update_costs(self, usage: dict):
        prompt_tokens = int(usage['prompt_tokens'])
        completion_tokens = int(usage['completion_tokens'])
        self._cost_manager.update_cost(prompt_tokens, completion_tokens, self.model)

    def get_costs(self) -> Costs:
        return self._cost_manager.get_costs()

    def get_max_tokens(self, messages: list[dict], functions: List[dict]=None):
        if not self.auto_max_tokens:
            return CONFIG.max_tokens_rsp
        return get_max_completion_tokens(messages, functions, self.model, CONFIG.max_tokens_rsp)

    def moderation(self, content: Union[str, list[str]]):
        try:
            if not content:
                logger.error("content cannot be empty!")
            else:
                rsp = self._moderation(content=content)
                return rsp
        except Exception as e:
            logger.error(f"moderating failed:{e}")

    def _moderation(self, content: Union[str, list[str]]):
        rsp = self.client.chat.moderations.create(input=content)
        return rsp

    async def amoderation(self, content: Union[str, list[str]]):
        try:
            if not content:
                logger.error("content cannot be empty!")
            else:
                rsp = await self._amoderation(content=content)
                return rsp
        except Exception as e:
            logger.error(f"moderating failed:{e}")

    async def _amoderation(self, content: Union[str, list[str]]):
        rsp = await self.aclient.chat.moderations.create(input=content)
        return rsp



import asyncio
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Field
from schema_agents.role import Role
from schema_agents.schema import Message
from schema_agents.tools.code_interpreter import create_mock_client
from schema_agents.utils.common import EventBus
from schema_agents.logs import logger
class MicroscopeControlRequirements(BaseModel):
    """Requirements for controlling the microscope and acquire images."""
    path: str = Field(default="", description="save images path")
    timeout: float = Field(default=0.0, description="timeout")
    query: str = Field(default="", description="Original request from the user")
    plan: str = Field(default="", description="plan for control microscope and acquiring images")

class MultiDimensionalAcquisitionScript(BaseModel):
    """Python script for simple and complex multi-dimensional acquisition.
    In the script, you can use the following functions to control the microscope:
    - `microscope_move({'x': 0.0, 'y': 0.0, 'z': 0.0})` # x, y, z are in microns
    - `microscope_snap({'path': './images', 'exposure': 0.0})` # path is the path to save the image, exposure is in seconds
    """
    script: str = Field(default="", description="Script for acquiring multi-dimensional images")
    explanation: str = Field(default="", description="Brief explanation for the script")
    timeout: float = Field(default=0.0, description="a reasonable timeout for executing the script")

class ExecutionResult(BaseModel):
    """Result of executing a Python script."""
    status: str = Field(description="Status of executing the script")
    outputs: List[Dict[str, Any]] = Field(default=[], description="Outputs of executing the script")
    traceback: Optional[str] = Field(default=None, description="Traceback of executing the script")

INIT_SCRIPT = """
def microscope_move(position):
    print(f"===> Moving to: {position}")

def microscope_snap(config):
    print(f"===> Snapped an image with exposure {config['exposure']} and saved to: { config['path']}")
"""


class Microscope():
    def __init__(self, client):
        self.client = client
        self.initialized = False

    async def plan(self, query: str=None, role: Role=None) -> MicroscopeControlRequirements:
        """Make a plan for image acquisition tasks."""
        return await role.aask(query, MicroscopeControlRequirements)
        
    async def multi_dimensional_acquisition(self, config: MicroscopeControlRequirements=None, role: Role=None) -> ExecutionResult:
        """Perform image acquisition by using Python script."""
        if not self.initialized:
            await self.client.executeScript({"script": INIT_SCRIPT})
            self.initialized = True
        print("Acquiring images in multiple dimensions: " + str(config))
        controlScript = await role.aask(config, MultiDimensionalAcquisitionScript)
        result = await self.client.executeScript({"script": controlScript.script, "timeout": controlScript.timeout})
        return ExecutionResult(
            status=result['status'],
            outputs=result['outputs'],
            traceback=result.get("traceback")
        )

def create_microscopist(client=None):
    if not client:
        client = create_mock_client()
    microscope = Microscope(client)
    microscopist = Role(
        name="Thomas",
        profile="Microscopist",
        goal="Acquire images from the microscope based on user's requests.",
        constraints=None,
        actions=[microscope.multi_dimensional_acquisition],
    )
    return microscopist

async def main():
    client = create_mock_client()
    event_bus = EventBus("microscopist")
    event_bus.register_default_events()

    microscope = Microscope(client)
    ms = Role(
        name="Thomas",
        profile="Microscopist",
        goal="Acquire images from the microscope based on user's requests.",
        constraints=None,
        actions=[microscope.plan, microscope.multi_dimensional_acquisition],
        event_bus=event_bus
    )

    messages = await ms.handle(Message(content="acquire an image and save to /tmp/img.png", role="User"))
    assert len(messages) == 2
    assert isinstance(messages[-1].data, ExecutionResult)
    assert messages[-1].data.status == "ok"

    messages = await ms.handle(Message(content="acquire image every 2nm along x, y in a 2x2um square, gradually increase exposure time from 0.1 to 2.0s", role="User"))
    assert len(messages) == 2
    assert isinstance(messages[-1].data, ExecutionResult)
    assert messages[-1].data.status == "ok"


    messages = await ms.handle(Message(content="acquire an image every 1 second for 10 seconds", role="User"))
    assert len(messages) == 2
    assert isinstance(messages[-1].data, ExecutionResult)
    assert messages[-1].data.status == "ok"

if __name__ == "__main__":
    asyncio.run(main())




from schema_agents.utils.schema_conversion import get_service_openapi_schema, get_service_json_schema, get_service_function_schema
from pydantic import BaseModel, Field
from typing import Optional, Dict, Any
import pytest

class DocWithScore(BaseModel):
    """A document with an associated relevance score."""
    doc: str = Field(description="The document retrieved.")
    score: float = Field(description="The relevance score of the retrieved document.")
    metadata: Dict[str, Any] = Field(description="The document's metadata.")
    base_url: Optional[str] = Field(
        None,
        description="The documentation's base URL, which will be used to resolve the relative URLs in the retrieved document chunks when producing markdown links.",
    )

def process(doc: DocWithScore) -> str:
    """Process a document with an associated relevance score."""
    return "processed"

def hello(name: str) -> str:
    """Say hello to someone."""
    return f"Hello, {name}!"

service_config = {
    "name": "Hello World",
    "id": "hello-world",
    "config": {
        "visibility": "public"
    },
    "process": process,
    "hello": {
        "world": hello
    }
}

expected_openapi_schema = {'openapi': '3.1.0', 'info': {'title': 'Hello World', 'version': 'v0.1.0'}, 'servers': [{'url': '/'}], 'paths': {'/process': {'post': {'description': 'Process a document with an associated relevance score.', 'operationId': 'process', 'requestBody': {'content': {'application/json': {'schema': {'$ref': '#/components/schemas/process'}}}, 'required': False}, 'responses': {'200': {'description': 'Successful response', 'content': {'application/json': {'schema': {'type': 'string'}}}}}, 'deprecated': False}}, '/hello.world': {'post': {'description': 'Say hello to someone.', 'operationId': 'hello_world', 'requestBody': {'content': {'application/json': {'schema': {'$ref': '#/components/schemas/hello_world'}}}, 'required': False}, 'responses': {'200': {'description': 'Successful response', 'content': {'application/json': {'schema': {'type': 'string'}}}}}, 'deprecated': False}}}, 'components': {'schemas': {'hello_world': {'properties': {'name': {'type': 'string', 'title': 'Name'}}, 'type': 'object', 'required': ['name'], 'title': 'hello_world', 'description': 'Say hello to someone.'}, 'DocWithScore': {'properties': {'doc': {'type': 'string', 'title': 'Doc', 'description': 'The document retrieved.'}, 'score': {'type': 'number', 'title': 'Score', 'description': 'The relevance score of the retrieved document.'}, 'metadata': {'type': 'object', 'title': 'Metadata', 'description': "The document's metadata."}, 'base_url': {'type': 'string', 'title': 'Base Url', 'description': "The documentation's base URL, which will be used to resolve the relative URLs in the retrieved document chunks when producing markdown links."}}, 'type': 'object', 'required': ['doc', 'score', 'metadata'], 'title': 'DocWithScore', 'description': 'A document with an associated relevance score.'}, 'process': {'properties': {'doc': {'allOf': [{'$ref': '#/components/schemas/DocWithScore'}], 'title': 'Doc', 'description': 'A document with an associated relevance score.'}}, 'type': 'object', 'required': ['doc'], 'title': 'process', 'description': 'Process a document with an associated relevance score.'}}}}

# Define the pytest test function
@pytest.mark.parametrize("field, expected_value", [
    ('openapi', '3.1.0'),
    ('info', {'title': 'Hello World', 'version': 'v0.1.0'}),
    ('servers', [{'url': '/'}]),
    ('paths', expected_openapi_schema['paths']),
    ('components', expected_openapi_schema['components'])
])
def test_validate_openapi_schema(field, expected_value):
    schema = get_service_openapi_schema(service_config)
    assert schema[field] == expected_value


expected_json_schema = {'process': {'input_schema': {'title': 'process', 'description': 'Process a document with an associated relevance score.', 'type': 'object', 'properties': {'doc': {'title': 'Doc', 'description': 'A document with an associated relevance score.', 'allOf': [{'$ref': '#/definitions/DocWithScore'}]}}, 'required': ['doc'], 'definitions': {'DocWithScore': {'title': 'DocWithScore', 'description': 'A document with an associated relevance score.', 'type': 'object', 'properties': {'doc': {'title': 'Doc', 'description': 'The document retrieved.', 'type': 'string'}, 'score': {'title': 'Score', 'description': 'The relevance score of the retrieved document.', 'type': 'number'}, 'metadata': {'title': 'Metadata', 'description': "The document's metadata.", 'type': 'object'}, 'base_url': {'title': 'Base Url', 'description': "The documentation's base URL, which will be used to resolve the relative URLs in the retrieved document chunks when producing markdown links.", 'type': 'string'}}, 'required': ['doc', 'score', 'metadata']}}}, 'output_schema': {'type': 'string'}}, 'hello.world': {'input_schema': {'title': 'hello_world', 'description': 'Say hello to someone.', 'type': 'object', 'properties': {'name': {'title': 'Name', 'type': 'string'}}, 'required': ['name']}, 'output_schema': {'type': 'string'}}}


@pytest.mark.parametrize("field, expected_value", [
    ('process', expected_json_schema['process']),
    ('hello.world', expected_json_schema['hello.world'])
])
def test_validate_json_schema(field, expected_value):
    schema = get_service_json_schema(service_config)
    assert schema[field] == expected_value
    

expected_function_schema = [{'type': 'function', 'function': {'name': 'process', 'description': 'Process a document with an associated relevance score.', 'parameters': {'type': 'object', 'properties': {'doc': {'allOf': [{'$ref': '#/definitions/DocWithScore'}]}}, 'required': ['doc'], 'definitions': {'DocWithScore': {'type': 'object', 'properties': {'doc': {'type': 'string'}, 'score': {'type': 'number'}, 'metadata': {'type': 'object'}, 'base_url': {'type': 'string'}}, 'required': ['doc', 'score', 'metadata']}}}}}, {'type': 'function', 'function': {'name': 'hello.world', 'description': 'Say hello to someone.', 'parameters': {'type': 'object', 'properties': {'name': {'type': 'string'}}, 'required': ['name']}}}]

@pytest.mark.parametrize("field, expected_value", [
    (0, expected_function_schema[0]),
    (1, expected_function_schema[1])
])
def test_validate_function_schema(field, expected_value):
    schema = get_service_function_schema(service_config)
    assert schema[field] == expected_value


import asyncio
import json
import traceback
from typing import List, Union

from schema_agents.role import Role
from schema_agents.schema import Message
from schema_agents.tools.code_interpreter import create_mock_client

from schema_agents.teams.image_analysis_hub.schemas import (PythonFunctionScript, PythonFunctionScriptChanges, Change,
                      PythonFunctionScriptWithLineNumber, SoftwareRequirement)

DEPLOY_SCRIPT = """
import asyncio
from imjoy_rpc.hypha import connect_to_server

{function_code}

if not 'hypha_server' in vars() or 'hypha_server' in globals():
    hypha_server = await connect_to_server(
        {{"name": "test client", "server_url": "https://ai.imjoy.io/", "method_timeout": 3000}}
    )

service_config = {{
    "name": "Hypha",
    "id": "{service_id}",
    "config": {{
        "visibility": "public"
    }},
}}

for function_name in {function_names}:
    service_config[function_name] = globals()[function_name]

await hypha_server.register_service(service_config)
"""


def add_line_number(script):
    return "\n".join(
        [f"{i+1:03d}→{line}" for i, line in enumerate(script.split("\n"))]
    )

def apply_changes(original_script: str, changes: List[Change]) -> str:
    original_script_lines = original_script.split("\n")
    line_offset = 0  # Offset to adjust line numbers after add/remove operations

    # Sort changes by line number
    sorted_changes = sorted(changes, key=lambda x: x.line)

    for change in sorted_changes:
        adjusted_line_number = (
            change.line - 1 + line_offset
        )  # Convert 1-based line number to 0-based index and adjust

        if change.operation == "remove":
            del original_script_lines[adjusted_line_number]
            line_offset -= 1  # Decrease offset due to removal
        elif change.operation == "add":
            original_script_lines.insert(adjusted_line_number, change.content)
            line_offset += 1  # Increase offset due to addition
        elif change.operation == "replace":
            original_script_lines[adjusted_line_number] = change.content

    updated_script = "\n".join(original_script_lines)
    return updated_script


async def fix_code(
    role: Role, client, python_function: PythonFunctionScript, output_summary
):
    prompt = f"Fix the code according to the error message:\n{output_summary}\n"
    "You should call the `PythonFunctionScriptChanges` function to make changes for fixing the error,"
    " optionally, if too many changes needed, call `PythonFunctionScript` to do a compete rewrite. DO NOT return text directly."
    "Importantly, DO NOT change the test script unless the error is caused by the test script."
    python_function_with_line_number = PythonFunctionScriptWithLineNumber(
        function_script_with_line_number=add_line_number(
            python_function.function_script
        ),
        pip_packages=python_function.pip_packages,
        test_script_with_line_number=add_line_number(python_function.test_script),
        docstring=python_function.docstring,
    )
    response = await role.aask(
        python_function_with_line_number,
        Union[PythonFunctionScriptChanges, PythonFunctionScript],
        prompt=prompt,
    )
    if isinstance(response, PythonFunctionScript):
        return response
    changes = response
    try:
        patched_test_script = (
            apply_changes(python_function.test_script, changes.test_script_changes)
            if changes.test_script_changes
            else python_function.test_script
        )
    except Exception:
        output_summary += (
            f"\nFailed to apply the test_script_changes:\n{changes.test_script_changes}\nTraceback: "
            + traceback.format_exc()
        )
        return await fix_code(role, client, python_function, output_summary)

    try:
        patched_function_script = (
            apply_changes(
                python_function.function_script, changes.function_script_changes
            )
            if changes.function_script_changes
            else python_function.function_script
        )
    except Exception:
        output_summary += (
            f"\nFailed to apply the function_script_changes:\n{changes.function_script_changes}\nTraceback: "
            + traceback.format_exc()
        )
        return await fix_code(role, client, python_function, output_summary)

    return PythonFunctionScript(
        function_names=python_function.function_names,
        docstring=python_function.docstring,
        function_script=patched_function_script,
        pip_packages=changes.updated_pip_packages,
        test_script=patched_test_script,
    )


async def generate_code(
    req: SoftwareRequirement, role: Role
) -> PythonFunctionScript:
    return await role.aask(req, PythonFunctionScript)

INSTALL_SCRIPT = """
try:
    import pyodide
    import micropip
    await micropip.install([{packages}])
except ImportError:
    import subprocess
    import sys
    subprocess.check_call([sys.executable, '-m', 'pip', 'install', {packages}])
"""

async def test_run_python_function(
    role, client, service_id, python_function: PythonFunctionScript
) -> PythonFunctionScript:
    """Test run the python function script."""
    if python_function.pip_packages:
        packages = ",".join([f"'{p}'" for p in python_function.pip_packages])
        results = await client.executeScript({"script": INSTALL_SCRIPT.format(packages=packages)})
        output_summary = json.dumps(
            {k: results[k] for k in results.keys() if results[k]}, indent=1
        )
        if results["status"] != "ok":
            raise RuntimeError(f"Failed to install pip packages: {python_function.pip_packages}, error: {output_summary}")
    results = await client.executeScript(
        {"script": python_function.function_script + "\n" + python_function.test_script}
    )
    assert results, "No results returned from the test script."
    if results["status"] != "ok":
        output_summary = json.dumps(
            {k: results[k] for k in results.keys() if results[k]}, indent=1
        )
        python_function = await fix_code(role, client, python_function, output_summary)
        return await test_run_python_function(role, client, service_id, python_function)
    else:
        # deploy the functions
        results = await client.executeScript(
            {"script": DEPLOY_SCRIPT.format(function_names=python_function.function_names,
                                 service_id=service_id,
                                 function_code=python_function.function_script
                                )
            }
        )
    return python_function

def create_data_engineer(client=None):
    async def develop_python_functions(
        req: SoftwareRequirement, role: Role
    ) -> PythonFunctionScript:
        """Develop python functions based on software requirements."""
        # if isinstance(req, SoftwareRequirement):
        func = await generate_code(req, role)
        try:
            func = await test_run_python_function(role, client, req.id, func)
        except RuntimeError as exp:
            req.additional_notes += f"\nPlease avoid the following error: {exp}"
            func = await generate_code(req, role)
            func = await test_run_python_function(role, client, req.id, func)
        return func
        # else:
        #     if client:
        #         await client.showDialog(
        #             src="https://gist.githubusercontent.com/oeway/b734c35f69a0ec0dcebe00b078676edb/raw/react-ui-plugin.imjoy.html",
        #             data={"jsx_script": req.jsx_script}
        #         )
        #     else:
        #         print("Unable to show the React UI")
        #     return func

    data_engineer = Role(
        name="Alice",
        profile="Data Engineer",
        goal="Develop the python function script according to the software requirement, ensuring that it fulfills the desired functionality. Implement necessary algorithms, handle data processing, and write tests to validate the correctness of the function.",
        constraints=None,
        actions=[develop_python_functions],
    )
    return data_engineer


async def main():
    mock_software_requirements = {
        "id": "cell_segmentation",
        "original_requirements": "Load an image file and segment the cells in the image, count the cells then show the result image. The cells are U2OS cells in a IF microscopy image, cells are round and in green color, the background is black. For testing, you can use /Users/wei.ouyang/workspace/LibreChat/chatbot/tests/data/img16.png",
        "python_function_requirements": [
            {
                "function_name": "load_image",
                "function_signature": "load_image(image_path: str) -> np.ndarray",
                "feature_instructions": "This function should load an image from the provided path and return it as a numpy array. Use the OpenCV library for this.",
                "context_requirements": "The function will be used to load an image from a file path.",
                "testing_requirements": "Test the function with different image file paths to ensure it can handle different image formats and sizes.",
            },
            {
                "function_name": "segment_cells",
                "function_signature": "segment_cells(image: np.ndarray) -> Tuple[np.ndarray, int]",
                "feature_instructions": "This function should segment the cells in the image and return the segmented image and the count of the cells. Use the OpenCV library for the segmentation. The cells are round and in green color, the background is black.",
                "context_requirements": "The function will be used to segment the cells in the image and count them.",
                "testing_requirements": "Test the function with different images to ensure it can accurately segment and count the cells.",
            },
        ],
        "react_ui_requirements": {
            "plugin_name": "CellCounter",
            "ui_layout": "The UI should have an input field to enter the image file path, a button to load and segment the image, and a display area to show the segmented image and the cell count.",
            "interaction_patterns": "The user enters the image file path and clicks the button to load and segment the image. The segmented image and the cell count are then displayed.",
            "functionalities": "The UI should allow the user to enter the image file path, load and segment the image, and display the segmented image and the cell count.",
            "user_flows": "The user enters the image file path, clicks the button to load and segment the image, and sees the segmented image and the cell count.",
        },
        "additional_notes": "The cells are U2OS cells in a IF microscopy image. The cells are round and in green color, the background is black. The number of cells in the image should be more than 12.",
    }

    DataEngineer = create_data_engineer(client=create_mock_client())
    ds = DataEngineer()

    pr = SoftwareRequirement.parse_obj(mock_software_requirements)
    req = Message(
        content=pr.json(),
        data=pr,
        role="Project Manager",
    )

    event_bus = ds.get_event_bus()
    event_bus.register_default_events()
    messages = await ds.handle(req)
    print(messages)


if __name__ == "__main__":
    asyncio.run(main())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   : the implement of serialization and deserialization

import copy
from typing import Tuple, Dict, List, Type
from pydantic import BaseModel, create_model, root_validator, validator
import pickle

from schema_agents.schema import Message, MemoryChunk


def actionoutout_schema_to_mapping(schema: Dict) -> Dict:
    """
    directly traverse the `properties` in the first level.
    schema structure likes
    ```
    {
        "title":"prd",
        "type":"object",
        "properties":{
            "Original Requirements":{
                "title":"Original Requirements",
                "type":"string"
            },
        },
        "required":[
            "Original Requirements",
        ]
    }
    ```
    """
    mapping = dict()
    for field, property in schema['properties'].items():
        if property['type'] == 'string':
            mapping[field] = (str, ...)
        elif property['type'] == 'array' and property['items']['type'] == 'string':
            mapping[field] = (List[str], ...)
        elif property['type'] == 'array' and property['items']['type'] == 'array':
            # here only consider the `Tuple[str, str]` situation
            mapping[field] = (List[Tuple[str, str]], ...)
    return mapping

class ActionOutput:
    content: str
    data: BaseModel

    def __init__(self, content: str, data: BaseModel):
        self.content = content
        self.data = data

    @classmethod
    def create_model_class(cls, class_name: str, mapping: Dict[str, Type]):
        new_class = create_model(class_name, **mapping)

        @validator('*', allow_reuse=True)
        def check_name(v, field):
            if field.name not in mapping.keys():
                raise ValueError(f'Unrecognized block: {field.name}')
            return v

        @root_validator(pre=True, allow_reuse=True)
        def check_missing_fields(values):
            required_fields = set(mapping.keys())
            missing_fields = required_fields - set(values.keys())
            if missing_fields:
                raise ValueError(f'Missing fields: {missing_fields}')
            return values

        new_class.__validator_check_name = classmethod(check_name)
        new_class.__root_validator_check_missing_fields = classmethod(check_missing_fields)
        return new_class

def serialize_message(message: Message):
    message_cp = copy.deepcopy(message)  # avoid `data` value update by reference
    ic = message_cp.data
    if ic:
        # model create by pydantic create_model like `pydantic.main.prd`, can't pickle.dump directly
        schema = ic.schema()
        mapping = actionoutout_schema_to_mapping(schema)

        message_cp.data = {
            'class': schema['title'],
            'mapping': mapping,
            'value': ic.dict()
        }
    msg_ser = pickle.dumps(message_cp)

    return msg_ser


def deserialize_message(message_ser: str) -> Message:
    message = pickle.loads(message_ser)
    if message.data:
        ic = message.data
        ic_obj = ActionOutput.create_model_class(class_name=ic['class'],
                                                 mapping=ic['mapping'])
        ic_new = ic_obj(**ic['value'])
        message.data = ic_new

    return message


def serialize_memory(memory: MemoryChunk):
    memory_cp = copy.deepcopy(memory)  # avoid `data` value update by reference
    ic = memory_cp.content
    if ic:
        # model create by pydantic create_model like `pydantic.main.prd`, can't pickle.dump directly
        schema = ic.schema()
        mapping = actionoutout_schema_to_mapping(schema)

        memory_cp.content = {
            'class': schema['title'],
            'mapping': mapping,
            'value': ic.dict()
        }
    mmr_ser = pickle.dumps(memory_cp)

    return mmr_ser


def deserialize_memory(mmr_ser: str) -> MemoryChunk:
    memory = pickle.loads(mmr_ser)
    if memory.content:
        ic = memory.content
        ic_obj = ActionOutput.create_model_class(class_name=ic['class'],
                                                 mapping=ic['mapping'])
        ic_new = ic_obj(**ic['value'])
        memory.content = ic_new

    return memory


"""Message schemas for message spec version 5"""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import re

from jsonschema import Draft4Validator, ValidationError

protocol_version = (5, 1)

# These fragments will be wrapped in the boilerplate for a valid JSON schema.
# We also add a default 'required' containing all keys.
schema_fragments: dict = {}


def get_msg_content_validator(msg_type, version_minor):
    frag = schema_fragments[msg_type]
    schema = {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "description": f"{msg_type} message contents schema",
        "type": "object",
        "properties": {},
        "additionalProperties": version_minor > protocol_version[1],
    }
    schema.update(frag)
    if "required" not in schema:
        # Require all keys by default
        schema["required"] = sorted(schema["properties"].keys())

    return Draft4Validator(schema)


header_part = {
    "type": "object",
    "properties": {
        "msg_id": {"type": "string"},
        "username": {"type": "string"},
        "session": {"type": "string"},
        # TODO - this is parsed to a datetime before we get it:
        "date": {},  # {"type": "string"},
        "msg_type": {"type": "string"},
        "version": {"type": "string"},
    },
    "required": ["msg_id", "username", "session", "date", "msg_type", "version"],
}

msg_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Jupyter message structure schema",
    "type": "object",
    "properties": {
        "header": header_part,
        "parent_header": {"type": "object"},
        "metadata": {"type": "object"},
        "content": {"type": "object"},  # Checked separately
        "buffers": {"type": "array"},
    },
    "required": ["header", "parent_header", "metadata", "content"],
}
msg_structure_validator = Draft4Validator(msg_schema)


def get_error_reply_validator(version_minor):
    return Draft4Validator(
        {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "description": "Jupyter 'error' reply schema",
            "type": "object",
            "properties": {
                "status": {"const": "error"},
                "ename": {"type": "string"},
                "evalue": {"type": "string"},
                "traceback": {"type": "array", "items": {"type": "string"}},
            },
            "required": ["status", "ename", "evalue", "traceback"],
            "additionalProperties": version_minor > protocol_version[1],
        }
    )


def get_abort_reply_validator(version_minor):
    return Draft4Validator(
        {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "description": "Jupyter 'abort' reply schema",
            "type": "object",
            "properties": {
                "status": {"const": "error"},
                "ename": {"type": "string"},
                "evalue": {"type": "string"},
                "traceback": {"type": "list", "items": {"type": "string"}},
            },
            "required": ["status", "ename", "evalue", "traceback"],
            "additionalProperties": version_minor > protocol_version[1],
        }
    )


reply_msgs_using_status = {
    "execute_reply",
    "inspect_reply",
    "complete_reply",
    "history_reply",
    "connect_reply",
    "comm_info_reply",
    "kernel_info_reply",
    "shutdown_reply",
    "interrupt_reply",
}


def validate_message(msg, msg_type=None, parent_id=None):  # noqa
    msg_structure_validator.validate(msg)

    msg_version_s = msg["header"]["version"]
    m = re.match(r"(\d+)\.(\d+)", msg_version_s)
    if not m:
        emsg = "Version {} not like 'x.y'"
        raise ValidationError(emsg)
    version_minor = int(m.group(2))

    if msg_type is not None:
        if msg["header"]["msg_type"] != msg_type:
            emsg = "Message type {!r} != {!r}".format(msg["header"]["msg_type"], msg_type)
            raise ValidationError(emsg)
    else:
        msg_type = msg["header"]["msg_type"]

    # Check for unexpected fields, unless it's a newer protocol version
    if version_minor <= protocol_version[1]:
        unx_top = set(msg) - set(msg_schema["properties"])
        if unx_top:
            msg = f"Unexpected keys: {unx_top}"
            raise ValidationError(msg)

        unx_header = set(msg["header"]) - set(header_part["properties"])
        if unx_header:
            emsg = f"Unexpected keys in header: {unx_header}"
            raise ValidationError(emsg)

    # Check the parent id
    if "reply" in msg_type and parent_id and msg["parent_header"]["msg_id"] != parent_id:
        emsg = "Parent header does not match expected"
        raise ValidationError(emsg)

    if msg_type in reply_msgs_using_status:
        # Most _reply messages have common 'error' and 'abort' structures
        try:
            status = msg["content"]["status"]
        except KeyError as e:
            raise ValidationError(str(e)) from None
        if status == "error":
            content_vdor = get_error_reply_validator(version_minor)
        elif status == "abort":
            content_vdor = get_abort_reply_validator(version_minor)
        elif status == "ok":
            content_vdor = get_msg_content_validator(msg_type, version_minor)
        else:
            msg = f"status {status!r} should be ok/error/abort"
            raise ValidationError(msg)
    else:
        content_vdor = get_msg_content_validator(msg_type, version_minor)

    content_vdor.validate(msg["content"])


# Shell messages ----------------------------------------------

schema_fragments["execute_request"] = {
    "properties": {
        "code": {"type": "string"},
        "silent": {"type": "boolean"},
        "store_history": {"type": "boolean"},
        "user_expressions": {"type": "object"},
        "allow_stdin": {"type": "boolean"},
        "stop_on_error": {"type": "boolean"},
    }
}

schema_fragments["execute_reply"] = {
    "properties": {
        # statuses 'error' and 'abort' change the structure, so check separately
        "status": {"const": "ok"},
        "execution_count": {"type": "number"},
        "payload": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {"source": {"type": "string"}},
                "additionalProperties": True,
            },
        },
        "user_expressions": {"type": "object"},
    },
    "required": ["status", "execution_count"],
}

schema_fragments["inspect_request"] = {
    "properties": {
        "code": {"type": "string"},
        "cursor_pos": {"type": "number"},
        "detail_level": {"enum": [0, 1]},
    }
}

schema_fragments["inspect_reply"] = {
    "properties": {
        # statuses 'error' and 'abort' change the structure, so check separately
        "status": {"const": "ok"},
        "found": {"type": "boolean"},
        "data": {"type": "object"},
        "metadata": {"type": "object"},
    }
}

schema_fragments["complete_request"] = {
    "properties": {
        "code": {"type": "string"},
        "cursor_pos": {"type": "number"},
    }
}

schema_fragments["complete_reply"] = {
    "properties": {
        # statuses 'error' and 'abort' change the structure, so check separately
        "status": {"const": "ok"},
        "matches": {"type": "array", "items": {"type": "string"}},
        "cursor_start": {"type": "number"},
        "cursor_end": {"type": "number"},
        "metadata": {"type": "object"},
    }
}

schema_fragments["history_request"] = {
    "properties": {
        "output": {"type": "boolean"},
        "raw": {"type": "boolean"},
        "hist_access_type": {"enum": ["range", "tail", "search"]},
        "session": {"type": "number"},
        "start": {"type": "number"},
        "stop": {"type": "number"},
        "n": {"type": "number"},
        "pattern": {"type": "string"},
        "unique": {"type": "boolean"},
    },
    "required": ["output", "raw", "hist_access_type"],
}

schema_fragments["history_reply"] = {
    "properties": {
        "status": {"const": "ok"},
        "history": {"type": "array", "items": {"minItems": 3, "maxItems": 3}},
    }
}

schema_fragments["is_complete_request"] = {
    "properties": {
        "code": {"type": "string"},
    }
}

schema_fragments["is_complete_reply"] = {
    "properties": {
        "status": {"enum": ["complete", "incomplete", "invalid", "unknown"]},
        "indent": {"type": "string"},
    },
    "required": ["status"],
}

# NB connect_request is deprecated
schema_fragments["connect_request"] = {"properties": {}}

schema_fragments["connect_reply"] = {
    "properties": {
        "shell_port": {"type": "number"},
        "iopub_port": {"type": "number"},
        "stdin_port": {"type": "number"},
        "hb_port": {"type": "number"},
        "control_port": {"type": "number"},
    }
}

schema_fragments["comm_info_request"] = {
    "properties": {
        "target_name": {"type": "string"},
    },
    "required": [],
}

schema_fragments["comm_info_reply"] = {
    "properties": {
        # statuses 'error' and 'abort' change the structure, so check separately
        "status": {"const": "ok"},
        "comms": {"type": "object"},
    }
}

schema_fragments["kernel_info_request"] = {"properties": {}}

schema_fragments["kernel_info_reply"] = {
    "properties": {
        # statuses 'error' and 'abort' change the structure, so check separately
        "status": {"const": "ok"},
        "protocol_version": {"type": "string"},
        "implementation": {"type": "string"},
        "implementation_version": {"type": "string"},
        "language_info": {"type": "object"},
        "banner": {"type": "string"},
        "debugger": {"type": "boolean"},
        "help_links": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {"text": {"type": "string"}, "url": {"type": "string"}},
            },
        },
    },
    "required": ["status", "protocol_version", "implementation", "language_info", "banner"],
}

schema_fragments["shutdown_request"] = {
    "properties": {
        "restart": {"type": "boolean"},
    }
}

schema_fragments["shutdown_reply"] = {
    "properties": {
        # statuses 'error' and 'abort' change the structure, so check separately
        "status": {"const": "ok"},
        "restart": {"type": "boolean"},
    }
}

schema_fragments["interrupt_request"] = {"properties": {}}
schema_fragments["interrupt_reply"] = {
    "properties": {
        # statuses 'error' and 'abort' change the structure, so check separately
        "status": {"const": "ok"},
    }
}

# IOPub messages ----------------------------------------------

mime_data = {
    "type": "object",
    "patternProperties": {r"^[\w\-\+\.]+/[\w\-\+\.]+$": {}},
    "additionalProperties": False,
}

schema_fragments["stream"] = {
    "properties": {
        "name": {"enum": ["stdout", "stderr"]},
        "text": {"type": "string"},
    }
}

schema_fragments["display_data"] = {
    "properties": {
        "data": mime_data,
        "metadata": {"type": "object"},
        "transient": {"type": "object"},
    },
    "required": ["data", "metadata"],
}

schema_fragments["update_display_data"] = {
    "properties": {
        "data": mime_data,
        "metadata": {"type": "object"},
        "transient": {"type": "object"},
    }
}

schema_fragments["execute_result"] = {
    "properties": {
        "execution_count": {"type": "number"},
        "data": mime_data,
        "metadata": {"type": "object"},
        "transient": {"type": "object"},
    },
    "required": ["execution_count", "data", "metadata"],
}

schema_fragments["clear_output"] = {
    "properties": {
        "wait": {"type": "boolean"},
    }
}

schema_fragments["execute_input"] = {
    "properties": {
        "code": {"type": "string"},
        "execution_count": {"type": "number"},
    }
}

schema_fragments["error"] = {
    "properties": {
        "ename": {"type": "string"},
        "evalue": {"type": "string"},
        "traceback": {"type": "array", "items": {"type": "string"}},
    }
}

schema_fragments["status"] = {
    "properties": {
        "execution_state": {"enum": ["busy", "idle", "starting"]},
    }
}

# Stdin messages ---------------------------------------------

schema_fragments["input_request"] = {
    "properties": {
        "prompt": {"type": "string"},
        "password": {"type": "number"},
    }
}

schema_fragments["input_reply"] = {
    "properties": {
        "value": {"type": "string"},
    }
}


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Time    : 2023/4/29 16:07
@Author  : alexanderwu
@File    : common.py
"""
import ast
import inspect
import os
import asyncio

from contextvars import ContextVar

current_session = ContextVar('current_session', default=None)

def check_cmd_exists(command) -> int:
    """Check if a command exists in the current system."""
    check_command = 'command -v ' + command + ' >/dev/null 2>&1 || { echo >&2 "no mermaid"; exit 1; }'
    result = os.system(check_command)
    return result


class NoMoneyException(Exception):
    """Raised when the operation cannot be completed due to insufficient funds"""

    def __init__(self, amount, message="Insufficient funds"):
        self.amount = amount
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f'{self.message} -> Amount required: {self.amount}'


def print_members(module, indent=0):
    """
    https://stackoverflow.com/questions/1796180/how-can-i-get-a-list-of-all-classes-within-current-module-in-python
    :param module:
    :param indent:
    :return:
    """
    prefix = ' ' * indent
    for name, obj in inspect.getmembers(module):
        print(name, obj)
        if inspect.isclass(obj):
            print(f'{prefix}Class: {name}')
            # print the methods within the class
            if name in ['__class__', '__base__']:
                continue
            print_members(obj, indent + 2)
        elif inspect.isfunction(obj):
            print(f'{prefix}Function: {name}')
        elif inspect.ismethod(obj):
            print(f'{prefix}Method: {name}')

class EventBus:
    """An event bus class."""

    def __init__(self, name, logger=None):
        """Initialize the event bus."""
        self._callbacks = {}
        self._logger = logger
        self.name = name

    def on(self, event_name, func):
        """Register an event callback."""
        if self._callbacks.get(event_name):
            self._callbacks[event_name].add(func)
        else:
            self._callbacks[event_name] = {func}
        return func

    def once(self, event_name, func):
        """Register an event callback that only run once."""
        if self._callbacks.get(event_name):
            self._callbacks[event_name].add(func)
        else:
            self._callbacks[event_name] = {func}
        # mark once callback
        self._callbacks[event_name].once = True
        return func

    async def aemit(self, event_name, *data):
        """Trigger an event."""
        futures = []
        for func in self._callbacks.get(event_name, []):
            try:
                if inspect.iscoroutinefunction(func):
                    futures.append(func(*data))
                else:
                    func(*data)
                if hasattr(func, "once"):
                    self.off(event_name, func)
            except Exception as e:
                if self._logger:
                    self._logger.error(
                        "Error in event callback: %s, %s, error: %s",
                        event_name,
                        func,
                        e,
                    )
        await asyncio.gather(*futures)

    def emit(self, event_name, *data):
        """Trigger an event."""
        for func in self._callbacks.get(event_name, []):
            try:
                if inspect.iscoroutinefunction(func):
                    asyncio.get_running_loop().create_task(func(*data))
                else:
                    func(*data)
                if hasattr(func, "once"):
                    self.off(event_name, func)
            except Exception as e:
                if self._logger:
                    self._logger.error(
                        "Error in event callback: %s, %s, error: %s",
                        event_name,
                        func,
                        e,
                    )

    def off(self, event_name, func=None):
        """Remove an event callback."""
        if not func:
            del self._callbacks[event_name]
        else:
            self._callbacks.get(event_name, []).remove(func)

    async def stream_callback(self, message):
        if message.type == "function_call":
            if message.status == "in_progress":
                print(message.arguments, end="")
            else:
                print(f'\nGenerating {message.name} ({message.status}): {message.arguments}')
        elif message.type == "text":
            print(message.content, end="")

    def register_default_events(self):
        self.on("stream", self.stream_callback)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Provide a configuration object that can be used to access the configuration
"""
import os
import openai

import yaml

from schema_agents.const import PROJECT_ROOT
from schema_agents.logs import logger
from schema_agents.utils.singleton import Singleton
# from schema_agents.tools import SearchEngineType, WebBrowserEngineType


class NotConfiguredException(Exception):
    """Exception raised for errors in the configuration.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message="The required configuration is not set"):
        self.message = message
        super().__init__(self.message)


class Config(metaclass=Singleton):
    """
    常规使用方法：
    config = Config("config.yaml")
    secret_key = config.get_key("MY_SECRET_KEY")
    print("Secret key:", secret_key)
    """

    _instance = None
    key_yaml_file = PROJECT_ROOT / "config/key.yaml"
    default_yaml_file = PROJECT_ROOT / "config/config.yaml"

    def __init__(self, yaml_file=default_yaml_file):
        self._configs = {}
        self._init_with_config_files_and_env(self._configs, yaml_file)
        logger.info("Config loading done.")
        self.global_proxy = self._get("GLOBAL_PROXY")
        self.openai_api_key = self._get("OPENAI_API_KEY")
        if not self.openai_api_key or "YOUR_API_KEY" == self.openai_api_key:
            raise NotConfiguredException("Set OPENAI_API_KEY first")

        self.openai_api_base = self._get("OPENAI_API_BASE")
        self.openai_proxy = None
        if not self.openai_api_base or "YOUR_API_BASE" == self.openai_api_base:
            self.openai_proxy = self._get("OPENAI_PROXY") or self.global_proxy
            if not self.openai_proxy:
                logger.info("Set OPENAI_API_BASE in case of network issues")

        self.openai_api_type = self._get("OPENAI_API_TYPE")
        self.openai_api_version = self._get("OPENAI_API_VERSION")
        self.openai_api_rpm = self._get("RPM", 3)
        self.openai_api_model = self._get("OPENAI_API_MODEL", "gpt-4-1106-preview")
        self.openai_seed = self._get("OPENAI_SEED", 42)
        self.openai_temperature = self._get("OPENAI_TEMPERATURE", 0)
        self.openai_stop = self._get("OPENAI_STOP")
        self.openai_logprobs = self._get("OPENAI_LOGPROBS")
        self.openai_top_logprobs = self._get("OPENAI_TOP_LOGPROBS")
        self.openai_timeout = self._get("OPENAI_TIMEOUT", 15)
        self.max_tokens_rsp = self._get("MAX_TOKENS", 2048)
        self.deployment_name = self._get("DEPLOYMENT_NAME")
        self.deployment_id = self._get("DEPLOYMENT_ID")

        self.serpapi_api_key = self._get("SERPAPI_API_KEY")
        self.serper_api_key = self._get("SERPER_API_KEY")
        self.google_api_key = self._get("GOOGLE_API_KEY")
        self.google_cse_id = self._get("GOOGLE_CSE_ID")
        # self.search_engine = self._get("SEARCH_ENGINE", SearchEngineType.SERPAPI_GOOGLE)
 
        # self.web_browser_engine = WebBrowserEngineType(self._get("WEB_BROWSER_ENGINE", "playwright"))
        # self.playwright_browser_type = self._get("PLAYWRIGHT_BROWSER_TYPE", "chromium")
        # self.selenium_browser_type = self._get("SELENIUM_BROWSER_TYPE", "chrome")
      
        self.long_term_memory = self._get('LONG_TERM_MEMORY', False)
        if self.long_term_memory:
            logger.warning("LONG_TERM_MEMORY is True")
        self.max_budget = self._get("MAX_BUDGET", 10.0)
        self.total_cost = 0.0

    def _init_with_config_files_and_env(self, configs: dict, yaml_file):
        """从config/key.yaml / config/config.yaml / env三处按优先级递减加载"""
        configs.update(os.environ)

        for _yaml_file in [yaml_file, self.key_yaml_file]:
            if not _yaml_file.exists():
                continue

            # 加载本地 YAML 文件
            with open(_yaml_file, "r", encoding="utf-8") as file:
                yaml_data = yaml.safe_load(file)
                if not yaml_data:
                    continue
                os.environ.update({k: v for k, v in yaml_data.items() if isinstance(v, str)})
                configs.update(yaml_data)

    def _get(self, *args, **kwargs):
        return self._configs.get(*args, **kwargs)

    def get(self, key, *args, **kwargs):
        """从config/key.yaml / config/config.yaml / env三处找值，找不到报错"""
        value = self._get(key, *args, **kwargs)
        if value is None:
            raise ValueError(f"Key '{key}' not found in environment variables or in the YAML file")
        return value


CONFIG = Config()



import asyncio
from functools import partial
from typing import Union
import yaml
from schema_agents.role import Role
from schema_agents.teams import Team

from schema_agents.teams.image_analysis_hub.data_engineer import create_data_engineer
from schema_agents.teams.image_analysis_hub.schemas import (GenerateUserForm, SoftwareRequirement, UserClarification,
                      UserRequirements)
from schema_agents.teams.image_analysis_hub.web_developer import ReactUI, create_web_developer
from schema_agents.teams.image_analysis_hub.microscopist import create_microscopist, MicroscopeControlRequirements

async def clarify_user_request(client, user_query: str, role: Role) -> Union[UserClarification, MicroscopeControlRequirements]:
    """Clarify user request by prompting to the user with a form."""
    config = await role.aask(user_query, Union[GenerateUserForm, MicroscopeControlRequirements])
    if isinstance(config, MicroscopeControlRequirements):
        return config
    fm = await client.createWindow(
        src="https://oeway.github.io/imjoy-json-schema-form/",
        config={
            "schema": config.form_schema and yaml.safe_load(config.form_schema),
            "ui_schema": config.ui_schema and yaml.safe_load(config.ui_schema),
            "submit_label": config.submit_label,
        },
        height="400px"
    )
    form = await fm.get_data()
    return UserClarification(user_query=user_query, form_data=str(form['formData']))

async def create_user_requirements(req: UserClarification, role: Role) -> Union[MicroscopeControlRequirements, UserRequirements]:
    """Respond to user's requests (can be control microscope or create software) after clarification."""
    return await role.aask(req, Union[MicroscopeControlRequirements, UserRequirements])

async def create_software_requirements(req: UserRequirements, role: Role) -> SoftwareRequirement:
    """Create software requirement."""
    return await role.aask(req, SoftwareRequirement)

async def deploy_app(ui: ReactUI, role: Role):
    """Deploy the app for sharing."""
    # serve_plugin(ui)
    print("Deploying the app...")

def create_image_analysis_hub(client, investment):
    """Create a team for image analysis."""
    team = Team(name="Image Analysis Hub", profile="A team of roles to create software for image analysis.", goal="Create a software for image analysis.", investment=investment)
    ux_manager = Role(name="Luisa",
        profile="UX Manager",
        goal="Focus on understanding the user's needs and experience. Understand the user needs by interacting with user and communicate these findings to the project manager by calling `UserRequirements`.",
        constraints=None,
        actions=[partial(clarify_user_request, client), create_user_requirements])

    project_manager = Role(name="Alice",
                profile="Project Manager",
                goal="Efficiently communicate with the user and translate the user's needs into software requirements",
                constraints=None,
                actions=[create_software_requirements])

    web_developer  = create_web_developer(client=client)
    data_engineer = create_data_engineer(client=client)
    # devops = Role(name="Bruce",
    #             profile="DevOps",
    #             goal="Deploy the software to the cloud and make it available to the user.",
    #             constraints=None,
    #             actions=[deploy_app])  
    
    # microscopist = create_microscopist(client=client)
    
    team.hire([ux_manager, project_manager, web_developer, data_engineer])
    return team



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Time    : 2023/4/29 15:50
@Author  : alexanderwu
@File    : __init__.py
"""

import os
import subprocess
import tempfile
import re
import json

from pydantic import BaseModel, create_model



def convert_key_name(key_name):
    words = key_name.split('_')
    capitalized_words = [word.capitalize() for word in words]
    return ' '.join(capitalized_words)

def dict_to_md(dict_obj):
    md_string = ""
    for key, value in dict_obj.items():
        md_string += f"\n## {convert_key_name(key)}\n\n"
        if isinstance(value, list):
            for item in value:
                if isinstance(item, tuple):
                    item = ', '.join(item)
                md_string += f"- {item}\n"
        else:
            md_string += f"{value}\n"
    return md_string

def apply_patch(original_text, patch_text):
    # Create a temporary file to hold the original code
    with tempfile.NamedTemporaryFile(mode='w+', delete=False) as original_file:
        original_file.write(original_text)
        original_path = original_file.name

    # Create a temporary file to hold the patch
    with tempfile.NamedTemporaryFile(mode='w+', delete=False) as patch_file:
        patch_file.write(patch_text)
        patch_path = patch_file.name

    # Use the patch command to apply the patch
    result = subprocess.run(['patch', original_path, patch_path], capture_output=True)

    # Read the patched content from the original file
    with open(original_path, 'r', encoding="utf-8") as file:
        patched_text = file.read()

    # Clean up the temporary files
    os.unlink(original_path)
    os.unlink(patch_path)

    if result.returncode != 0:
        raise RuntimeError(f"Failed to apply patch: {result.stdout and result.stdout.decode()}\n{result.stderr and result.stderr.decode()}")
    else:
        return patched_text


def parse_special_json(json_string):
    # Regex pattern to find string values enclosed in double quotes or backticks, considering escaped quotes
    pattern = r'"(?:[^"\\]|\\.)*"|`[^`]*`'
    # Extract all matches and store them in a list
    code_blocks = re.findall(pattern, json_string)

    mapping = {}
    # Replace each match in the JSON string with a special placeholder
    for i, block in enumerate(code_blocks):
        json_string = json_string.replace(f'{block}', f'"###CODE-BLOCK-PLACEHOLDER-{i}###"')
        mapping[f'###CODE-BLOCK-PLACEHOLDER-{i}###'] = block[1:-1].encode('utf-8').decode('unicode_escape')

    # Parse the JSON string into a Python dictionary
    data = json.loads(json_string)
    
    def restore_codeblock(data):
        if isinstance(data, str):
            if re.match(r'###CODE-BLOCK-PLACEHOLDER-\d+###', data):
                return mapping[data]
            else:
                return data
        if isinstance(data, (int, float, bool)) or data is None:
            return data
        # Replace each placeholder with the corresponding code block
        if isinstance(data, list):
            cdata = []
            for d in data:
                cdata.append(restore_codeblock(d))
            return cdata

        assert isinstance(data, dict)
        cdata = {}
        for key in list(data.keys()):
            value = data[key]
            value = restore_codeblock(value)
            key = restore_codeblock(key)
            cdata[key] = value
        return cdata
    
    return restore_codeblock(data)

# https://stackoverflow.com/a/58938747
def remove_a_key(d, remove_key):
    if isinstance(d, dict):
        for key in list(d.keys()):
            if key == remove_key:
                del d[key]
            else:
                remove_a_key(d[key], remove_key)

def schema_to_function(schema: BaseModel):
    assert schema.__doc__, f"{schema.__name__} is missing a docstring."
    assert (
        "title" not in schema.__fields__.keys()
    ), "`title` is a reserved keyword and cannot be used as a field name."
    schema_dict = schema.schema()
    remove_a_key(schema_dict, "title")
    remove_a_key(schema_dict, "description")

    return {
        "name": schema.__name__,
        "description": schema.__doc__,
        "parameters": schema_dict,
    }


def dict_to_pydantic_model(name: str, dict_def: dict, doc: str = None):
    fields = {}
    for field_name, value in dict_def.items():
        if isinstance(value, tuple):
            fields[field_name] = value
        elif isinstance(value, dict):
            fields[field_name] = (dict_to_pydantic_model(f"{name}_{field_name}", value), ...)
        else:
            raise ValueError(f"Field {field_name}:{value} has invalid syntax")
    model = create_model(name, **fields)
    model.__doc__ = doc
    return model



"""Provide conversion functions for OpenAPI, JSON, and function schemas."""
from pydantic import BaseModel, Field
from inspect import signature
import inspect

from openapi_schema_pydantic import OpenAPI
from openapi_schema_pydantic.util import PydanticSchema, construct_open_api_with_schema_class
from pydantic import BaseModel, Field
from schema_agents.utils import schema_to_function, dict_to_pydantic_model

def extract_schemas(func, func_name=None):
    assert callable(func), "Tools must be callable functions"
    sig = signature(func)
    # var_positional = [
    #     p.name for p in sig.parameters.values() if p.kind == p.VAR_POSITIONAL
    # ]
    # kwargs_args = [
    #     p.name for p in sig.parameters.values() if p.kind != p.VAR_POSITIONAL
    # ]
    names = [p.name for p in sig.parameters.values()]
    for name in names:
        assert sig.parameters[name].annotation != inspect._empty, f"Parameter {name} must have type annotation"
    types = [sig.parameters[name].annotation for name in names]
    defaults = []
    for i, name in enumerate(names):
        if sig.parameters[name].default == inspect._empty:
            # if types[i] is not pydantic base model
            if not isinstance(types[i], type) or not issubclass(types[i], BaseModel):
                defaults.append(Field(...))
            else: 
                defaults.append(Field(..., description=types[i].__doc__))
        else:
            defaults.append(
                Field(
                    sig.parameters[name].default, description=types[i].__doc__
                )
            )
    
    
    func_name = func_name or func.__name__
    return (
        dict_to_pydantic_model(
            func_name,
            {names[i]: (types[i], defaults[i]) for i in range(len(names))},
            func.__doc__,
        ),
        sig.return_annotation,
    )


def get_primitive_schema(type_, is_json_schema=False):
    """Maps Python types to OpenAPI schema types."""
    if type_ is str:
        return {"type": "string"}
    elif type_ is int:
        return {"type": "integer"}
    elif type_ is float:
        return {"type": "number"}
    elif type_ is bool:
        return {"type": "boolean"}
    elif is_json_schema:
        if type_ is inspect._empty:
            return {"type": "null"}
        elif inspect.isclass(type_) and issubclass(type_, BaseModel):
            return type_.schema()
        else:
            raise ValueError(f"Unsupported type: {type_}")
    else:
        if type_ is inspect._empty:
            return {}
        elif inspect.isclass(type_) and issubclass(type_, BaseModel):
            return PydanticSchema(schema_class=type_)
        else:
            raise ValueError(f"Unsupported type: {type_}")

def create_function_openapi_schema(func, func_name=None, method="post"):
    func_name = func_name or func.__name__
    input_schema, output_schema = extract_schemas(func, func_name=func_name)
    output_schema_type = get_primitive_schema(output_schema, is_json_schema=False)
    return {
        method: {
            "description": func.__doc__,
            "operationId": func_name,
            "requestBody": {
                "content": {"application/json": {
                    "schema": PydanticSchema(schema_class=input_schema)
                }},
            },
            "responses": {"200": {
                "description": "Successful response",
                "content": {"application/json": {
                    "schema": output_schema_type if output_schema_type else None
                }},
            }},
        }
    }

def create_function_json_schema(func, func_name=None):
    func_name = func_name or func.__name__
    input_schema, output_schema = extract_schemas(func, func_name=func_name)
    output_schema_type = get_primitive_schema(output_schema, is_json_schema=True)
    return input_schema.schema(), output_schema_type

def get_service_functions(service_config):
    functions = {}
    
    def extract_functions(config, path=""):
        for key, value in config.items():
            if isinstance(value, dict):
                extract_functions(value, path + key + ".")
            elif callable(value):
                functions[path + key] = value
    
    extract_functions(service_config)
    return functions

def get_service_openapi_schema(service_config, service_url="/"):
    functions = get_service_functions(service_config)
    paths = {}
    for path, func in functions.items():
        paths[f"/{path}"] = create_function_openapi_schema(func, func_name=path.replace(".", "_"))
    
    open_api = OpenAPI.parse_obj({
        "info": {"title": service_config['name'], "version": "v0.1.0"},
        "servers": [{"url": service_url or "/"}],
        "paths": paths,
    })
    open_api = construct_open_api_with_schema_class(open_api)

    # Return the generated OpenAPI schema in JSON format
    return open_api.dict(by_alias=True, exclude_none=True)


def get_service_json_schema(service_config):
    functions = get_service_functions(service_config)
    schemas = {}

    for path, func in functions.items():
        input_schema, output_schema = create_function_json_schema(func, func_name=path.replace(".", "_"))
        schemas[path] = {
            "input_schema": input_schema,
            "output_schema": output_schema
        }
    return schemas


def get_service_function_schema(service_config):
    functions = get_service_functions(service_config)
    function_schemas = []

    for path, func in functions.items():
        input_schema, _ = extract_schemas(func, func_name=path)
        function_schemas.append({"type": "function", "function": schema_to_function(input_schema)})
    return function_schemas



import asyncio

from schema_agents.role import Role
from schema_agents.schema import Message
from schema_agents.tools.code_interpreter import create_mock_client

from schema_agents.teams.image_analysis_hub.schemas import ReactUI, SoftwareRequirement

def create_web_developer(client=None):

    async def develop_react_ui(req: SoftwareRequirement, role: Role) -> ReactUI:
        """Develop the React UI plugin according to the software requirement, ensuring that it fulfills the desired functionality. """
        plugin = await role.aask(req, ReactUI)
        if client:
            await client.createWindow(
                src="https://gist.githubusercontent.com/oeway/b734c35f69a0ec0dcebe00b078676edb/raw/react-ui-plugin.imjoy.html",
                data={"jsx_script": plugin.jsx_script, "service_id": req.id}
            )
        return plugin

    web_developer = Role(name="Bob",
                                    profile="Web Developer",
                                    goal="Develop the React UI plugin according to the software requirement, ensuring that it fulfills the desired functionality. Implement necessary algorithms, handle data processing, and write tests to validate the correctness of the function.",
                                    constraints=None,
                                    actions=[develop_react_ui])
    return web_developer

def serve_plugin(plugin: ReactUI):
    import uvicorn
    from fastapi import FastAPI
    from fastapi.responses import HTMLResponse
    from jinja2 import Environment, PackageLoader, select_autoescape

    jinja_env = Environment(
        loader=PackageLoader("schema_agents"), autoescape=select_autoescape()
    )

    app = FastAPI(title="web app")

    @app.get("/", response_class=HTMLResponse)
    async def index():
        temp = jinja_env.get_template("react_template.html")
        source = temp.render(plugin.dict())
        return source

    app.mount("/", index)
    uvicorn.run(app, host="127.0.0.1", port=8000)

async def main(context=None):
    mock_software_requirements = {
    "id": "cell_counter",
    "original_requirements": "Load an image file and segment the cells in the image, count the cells then show the result image. The cells are U2OS cells in a IF microscopy image, cells are round and in green color, the background is black.",
    "python_function_requirements": [
        {
        "function_name": "load_image",
        "function_signature": "load_image(image_path: str) -> np.ndarray",
        "feature_instructions": "This function should load an image from the provided path and return it as a numpy array. Use the OpenCV library for this.",
        "context_requirements": "The function will be used to load an image from a file path.",
        "testing_requirements": "Test the function with different image file paths to ensure it can handle different image formats and sizes."
        },
        {
        "function_name": "segment_cells",
        "function_signature": "segment_cells(image: np.ndarray) -> Tuple[np.ndarray, int]",
        "feature_instructions": "This function should segment the cells in the image and return the segmented image and the count of the cells. Use the OpenCV library for the segmentation. The cells are round and in green color, the background is black.",
        "context_requirements": "The function will be used to segment the cells in the image and count them.",
        "testing_requirements": "Test the function with different images to ensure it can accurately segment and count the cells."
        }
    ],
    "react_ui_requirements": {
        "plugin_name": "CellCounter",
        "ui_layout": "The UI should have an input field to enter the image file path, a button to load and segment the image, and a display area to show the segmented image and the cell count.",
        "interaction_patterns": "The user enters the image file path and clicks the button to load and segment the image. The segmented image and the cell count are then displayed.",
        "functionalities": "The UI should allow the user to enter the image file path, load and segment the image, and display the segmented image and the cell count.",
        "user_flows": "The user enters the image file path, clicks the button to load and segment the image, and sees the segmented image and the cell count."
    },
    "additional_notes": "The cells are U2OS cells in a IF microscopy image. The cells are round and in green color, the background is black. The number of cells in the image should be more than 12."
    }
    
    mock_plugin = {
        "plugin_name": "CellCounter",
        "root_element": "root",
        "react_version": "17.0.2",
        "react_dom_version": "17.0.2",
        "babel_version": "6.26.0",
        "jsx_script": """
            class CellCounter extends React.Component {
            constructor(props) {
                super(props);
                this.state = {
                imagePath: '',
                segmentedImage: null,
                cellCount: 0
                };
                this.handleImagePathChange = this.handleImagePathChange.bind(this);
                this.handleLoadAndSegment = this.handleLoadAndSegment.bind(this);
            }

            handleImagePathChange(event) {
                this.setState({imagePath: event.target.value});
            }

            async handleLoadAndSegment() {
                const image = await window.python.load_image(this.state.imagePath);
                const [segmentedImage, cellCount] = await window.python.segment_cells(image);
                this.setState({
                segmentedImage: segmentedImage,
                cellCount: cellCount
                });
            }

            render() {
                return (
                <div>
                    <input type='text' value={this.state.imagePath} onChange={this.handleImagePathChange} />
                    <button onClick={this.handleLoadAndSegment}>Load and Segment Image</button>
                    {this.state.segmentedImage && <img src={this.state.segmentedImage} />}
                    <p>Cell Count: {this.state.cellCount}</p>
                </div>
                );
            }
            }

            ReactDOM.render(
            <CellCounter />,
            document.getElementById('root')
            );
        """
    }
    # context['plugin'] = ReactUI.parse_obj(mock_plugin)
    WebDeveloper = create_web_developer(client=create_mock_client())
    wd = WebDeveloper()
    pr = SoftwareRequirement.parse_obj(mock_software_requirements)
    req = Message(content=pr.json(), data=pr, role="Project Manager")
    resp = await wd.handle(req)
    print(resp)
    if context is not None:
        context['plugin'] = resp[0].data
    

if __name__ == "__main__":
    context = {}
    asyncio.run(main(context))
    serve_plugin(context['plugin'])


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Time    : 2023/5/5 23:04
@Author  : alexanderwu
@File    : base_gpt_api.py
"""
from abc import abstractmethod
from typing import Optional, Union, Dict, List, Any

from schema_agents.logs import logger
from schema_agents.provider.base_chatbot import BaseChatbot
from schema_agents.utils.common import EventBus


class BaseGPTAPI(BaseChatbot):
    """GPT API abstract class, requiring all inheritors to provide a series of standard capabilities"""
    system_prompt = 'You are a helpful assistant.'

    def _user_msg(self, msg: str) -> dict[str, str]:
        if isinstance(msg, str):
            return {"role": "user", "content": msg}
        else:
            raise TypeError(f"msg must be str or dict, but got {type(msg)}")

    def _assistant_msg(self, msg: str) -> dict[str, str]:
        return {"role": "assistant", "content": msg}

    def _system_msg(self, msg: str) -> dict[str, str]:
        return {"role": "system", "content": msg}

    def _system_msgs(self, msgs: list[str]) -> list[dict[str, str]]:
        return [self._system_msg(msg) for msg in msgs]

    def _default_system_msg(self):
        return self._system_msg(self.system_prompt)

    def ask(self, msg: Union[str, Dict[str, str]], functions: List[Dict[str, Any]]=None, function_call: Union[str, Dict[str, str]]=None, event_bus:EventBus=None) -> str:
        message = [self._default_system_msg(), self._user_msg(msg)]
        if function_call is not None:
            assert isinstance(function_call, dict) or function_call in ['none', 'auto'], f"function_call must be dict or 'none', 'auto', but got {function_call}"
        rsp = self.completion(message, functions=functions, function_call=function_call, event_bus=event_bus)
        fc = rsp.get("choices")[0]["message"].get("function_call")
        if fc is not None:
            return rsp.get("choices")[0]["message"].get("function_call")
        return self.get_choice_text(rsp)

    async def aask(self, msg: Union[str, Dict[str, str], List[Dict[str, str]]], system_msgs: Optional[list[str]] = None, functions: List[Dict[str, Any]]=None, function_call: Union[str, Dict[str, str]]=None, event_bus: EventBus=None, use_tool_calls=True) -> str:
        if isinstance(msg, list):
            messages = []
            for m in msg:
                if isinstance(m, str):
                    messages.append(self._user_msg(m))
                else:
                    messages.append(m)
            if system_msgs:
                messages = self._system_msgs(system_msgs) + messages
        else:
            user_msg = self._user_msg(msg) if isinstance(msg, str) else msg
            if system_msgs:
                messages = self._system_msgs(system_msgs) + [user_msg]
            else:
                messages = [self._default_system_msg(), user_msg]
        if function_call is not None:
            assert isinstance(function_call, dict) or function_call in ['none', 'auto'], f"function_call must be dict or 'none', 'auto', but got {function_call}"
        if functions:
            f_names = [f["name"] for f in functions]
            assert len(functions) == len(set(f_names)), f"functions must have unique names, but got {f_names}"
            if use_tool_calls:
                if isinstance(function_call, dict):
                    tool_choice = {"type": "function", "function": function_call}
                else:
                    tool_choice = function_call
                rsp = await self.acompletion_tool(messages, tools=[{"type": "function", "function": func} for func in functions], tool_choice=tool_choice, event_bus=event_bus)
            else:
                rsp = await self.acompletion_function(messages, functions=functions, function_call=function_call, event_bus=event_bus)
        else:
            rsp = await self.acompletion_text(messages, stream=True, event_bus=event_bus)
        # logger.debug(message)
        logger.debug(rsp)
        return rsp

    def _extract_assistant_rsp(self, context):
        return "\n".join([i["content"] for i in context if i["role"] == "assistant"])

    def ask_batch(self, msgs: list, event_bus:EventBus=None) -> str:
        context = []
        for msg in msgs:
            umsg = self._user_msg(msg)
            context.append(umsg)
            rsp = self.completion(context)
            rsp_text = self.get_choice_text(rsp)
            context.append(self._assistant_msg(rsp_text))
        return self._extract_assistant_rsp(context, event_bus=event_bus)

    async def aask_batch(self, msgs: list, event_bus:EventBus=None) -> str:
        """Sequential questioning"""
        context = []
        for msg in msgs:
            umsg = self._user_msg(msg)
            context.append(umsg)
            rsp_text = await self.acompletion_text(context, event_bus=event_bus)
            context.append(self._assistant_msg(rsp_text))
        return self._extract_assistant_rsp(context)

    def ask_code(self, msgs: list[str], event_bus:EventBus=None) -> str:
        """FIXME: No code segment filtering has been done here, and all results are actually displayed"""
        rsp_text = self.ask_batch(msgs, event_bus=event_bus)
        return rsp_text

    async def aask_code(self, msgs: list[str], event_bus:EventBus=None) -> str:
        """FIXME: No code segment filtering has been done here, and all results are actually displayed"""
        rsp_text = await self.aask_batch(msgs, event_bus=event_bus)
        return rsp_text

    @abstractmethod
    def completion(self, messages: list[dict], event_bus:EventBus=None):
        """All GPTAPIs are required to provide the standard OpenAI completion interface
        [
            {"role": "system", "content": "You are a helpful assistant."},
            {"role": "user", "content": "hello, show me python hello world code"},
            # {"role": "assistant", "content": ...}, # If there is an answer in the history, also include it
        ]
        """

    @abstractmethod
    async def acompletion(self, messages: list[dict], event_bus:EventBus=None):
        """Asynchronous version of completion
        All GPTAPIs are required to provide the standard OpenAI completion interface
        [
            {"role": "system", "content": "You are a helpful assistant."},
            {"role": "user", "content": "hello, show me python hello world code"},
            # {"role": "assistant", "content": ...}, # If there is an answer in the history, also include it
        ]
        """

    @abstractmethod
    async def acompletion_text(self, messages: list[dict], stream=False, event_bus:EventBus=None) -> str:
        """Asynchronous version of completion. Return str. Support stream-print"""

    def get_choice_text(self, rsp: dict) -> str:
        """Required to provide the first text of choice"""
        return rsp.choices[0].dict()["message"]["content"]
    
    def get_choice_logprobs(self, rsp: dict) -> str:
        """Required to provide the first logprobs of choice"""
        return rsp.choices[0].dict()["logprobs"]["content"]

    def messages_to_prompt(self, messages: list[dict]):
        """[{"role": "user", "content": msg}] to user: <msg> etc."""
        return '\n'.join([f"{i['role']}: {i['content']}" for i in messages])

    def messages_to_dict(self, messages):
        """objects to [{"role": "user", "content": msg}] etc."""
        return [i.to_dict() for i in messages]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pytest
from typing import List, Optional, Union
from pydantic import BaseModel, Field
from schema_agents.role import Role
from schema_agents.schema import Message
import json


class SoftwareRequirementDocument(BaseModel):
    """Write Software Requirement Document."""
    original_requirements: str = Field(description="The polished complete original requirements")
    product_goals: List[str] = Field(description="Up to 3 clear, orthogonal product goals. If the requirement itself is simple, the goal should also be simple")
    user_stories: List[str] = Field(description="up to 5 scenario-based user stories, If the requirement itself is simple, the user stories should also be less")
    ui_design_draft: str = Field(description="Describe the elements and functions, also provide a simple style description and layout description.")
    anything_unclear: str = Field(None, description="Make clear here.")

class UserClarification(BaseModel):
    """Provide more details for the use case."""
    content: str = Field(description="Anwser to the clarification request.")

class SearchInternetQuery(BaseModel):
    """Keywords for searching the internet."""
    query: str = Field(description="space separated keywords for searching the internet.")

class GetExtraInformation(BaseModel):
    """Extra information needed to be able to work on the task."""
    content: str = Field(description="The information.")
    # summary: str = Field(description="Summary of what you already get.")

async def search_internet(query: str) -> str:
    """Search internet for more information."""
    return "Nothing found"

class UserInput(BaseModel):
    """User clarification"""
    goal: str = Field(description="Clarified goal")
    notes: str = Field(description="Additional notes")

async def get_user_input(query: str) -> str:
    """Get additional information from user."""
    return UserInput(
        goal="The goal is to get the image of the green cells under IF staining and count the number of cells in the image.",
        notes="The software should accept user uploaded files, has a simple web interface, the code should meet PEP8 standard, the rest can be decided by the developer."
    )

async def clarify(query: GetExtraInformation, role: Role) -> UserClarification:
    """Clarify the requirements."""
    return await role.aask(query, UserClarification)



class FormDialogInfo(BaseModel):
    """Create a JSON Schema Form Dialog using react-jsonschema-form to get more information from the user.
    Whenever possible, try to propose the options for the user to choose from, instead of asking the user to type in the text."""
    form_schema: str = Field(description="json schema for the fields, in yaml format")
    ui_schema: Optional[str] = Field(None, description="customized ui schema for rendering the form, json string, no need to escape quotes, in yaml format")
    submit_label: Optional[str] = Field("Submit", description="Submit button label")


@pytest.mark.asyncio
async def test_schema_user():
    def get_user_response(config: FormDialogInfo) -> UserClarification:
        """Get user response."""
        return UserClarification(content=str({"anwser": "I don't know"}))
    user = Role(name="Bob",
        profile="User",
        goal="Provide the use case and requirements for the development, the aim is to create a cell counting software for U2OS cells under confocal microscope, I would like to use web interface in Imjoy.",
        constraints=None,
        actions=[get_user_response])
    user.get_event_bus().register_default_events()
    # create a form_schema for get user name
    form_schema = json.dumps({"title": "Get User Name", "type": "object", "properties": {"name": {"type": "string"}}})
    form_dialog = FormDialogInfo(form_schema=form_schema)
    msg = Message(content=form_dialog.json(), data=form_dialog, role="Boss")
    responses = await user.handle(msg)
    assert isinstance(responses[0].data, UserClarification)

@pytest.mark.asyncio
async def test_str_streaming():
    async def create_user_requirements(query: str, role: Role) -> str:
        """Create user requirements."""
        return await role.aask(query, str)

    bioimage_analyst = Role(name="Alice",
                profile="BioImage Analyst",
                goal="Efficiently communicate with the user and translate the user's needs into software requirements",
                constraints=None,
                actions=[create_user_requirements])
    bioimage_analyst.get_event_bus().register_default_events()
    responses = await bioimage_analyst.handle(Message(role="Bot", content="What can you do for me?"))
    assert isinstance(responses[0].content, str)


@pytest.mark.asyncio
async def test_tool_call():
    async def create_user_requirements(query: str, role: Role) -> SoftwareRequirementDocument:
        """Create user requirements."""
        def create_sdr(sdr: SoftwareRequirementDocument) -> str:
            """Create Software Requirement Document."""
            return "SDR"

        async def get_extra_information(extra_info: GetExtraInformation, hint: str) -> UserInput:
            """Get Extra Information with hint to the user."""
            return await get_user_input(extra_info)

        response = await role.acall(query, [create_sdr, get_extra_information], SoftwareRequirementDocument)
        return response
        
    bioimage_analyst = Role(name="Alice",
                profile="BioImage Analyst",
                goal="Efficiently communicate with the user and translate the user's needs into software requirements",
                constraints=None,
                actions=[create_user_requirements])
    bioimage_analyst.get_event_bus().register_default_events()
    responses = await bioimage_analyst.handle(Message(role="Bot", content="Create a segmentation software"))
    assert isinstance(responses[0].data, SoftwareRequirementDocument)


@pytest.mark.asyncio
async def test_parallel_function_call():
    async def create_user_requirements(query: str, role: Role) -> SoftwareRequirementDocument:
        """Create user requirements."""
        response = await role.aask(query, Union[SoftwareRequirementDocument, GetExtraInformation], use_tool_calls=True)
        if isinstance(response, SoftwareRequirementDocument):
            return response
        elif isinstance(response, GetExtraInformation):
            user_req = await get_user_input(response)
            return await role.aask(user_req, SoftwareRequirementDocument)
        else:
            raise TypeError(f"response must be SoftwareRequirementDocument or GetExtraInformation, but got {type(response)}")

    bioimage_analyst = Role(name="Alice",
                profile="BioImage Analyst",
                goal="Efficiently communicate with the user and translate the user's needs into software requirements",
                constraints=None,
                actions=[create_user_requirements])
    bioimage_analyst.get_event_bus().register_default_events()
    responses = await bioimage_analyst.handle(Message(role="Bot", content="Create a segmentation software"))
    assert isinstance(responses[0].data, SoftwareRequirementDocument)

@pytest.mark.asyncio
async def test_schema_str_input():
    async def create_user_requirements(query: str, role: Role) -> SoftwareRequirementDocument:
        """Create user requirements."""
        response = await role.aask(query, Union[SoftwareRequirementDocument, GetExtraInformation])
        if isinstance(response, SoftwareRequirementDocument):
            return response
        elif isinstance(response, GetExtraInformation):
            user_req = await get_user_input(response)
            return await role.aask(user_req, SoftwareRequirementDocument)
        else:
            raise TypeError(f"response must be SoftwareRequirementDocument or GetExtraInformation, but got {type(response)}")

    bioimage_analyst = Role(name="Alice",
                profile="BioImage Analyst",
                goal="Efficiently communicate with the user and translate the user's needs into software requirements",
                constraints=None,
                actions=[create_user_requirements])
    bioimage_analyst.get_event_bus().register_default_events()
    responses = await bioimage_analyst.handle(Message(role="Bot", content="Create a segmentation software"))
    assert isinstance(responses[0].data, SoftwareRequirementDocument)

@pytest.mark.asyncio
async def test_schemas():
    user = Role(name="Bob",
                profile="User",
                goal="Provide the use case and requirements for the development, the aim is to create a cell counting software for U2OS cells under confocal microscope, I would like to use web interface in Imjoy.",
                constraints=None,
                actions=[clarify, search_internet])
    user.get_event_bus().register_default_events()
    responses = await user.handle(Message(role="Bot", content="Find more information on the internet about cell counting software."))
    assert responses[0].data is None
    assert responses[0].content == "Nothing found"

    instruct = GetExtraInformation(content="Tell me the use case in 1 sentence.", summary="Requesting details about the use case")
    responses = await user.handle(Message(role="Bot", content=instruct.json(), data=instruct))
    assert isinstance(responses[0].data, UserClarification)



@pytest.mark.asyncio
async def test_schema_str():
    async def process_user_input(query: GetExtraInformation, role: Role) -> str:
        """Process user input."""
        query = await role.aask(query, SearchInternetQuery)
        return query

    user = Role(name="Bob",
                profile="User",
                goal="Provide the use case and requirements for the development, the aim is to create a cell counting software for U2OS cells under confocal microscope, I would like to use web interface in Imjoy.",
                constraints=None,
                actions=[process_user_input])
    user.get_event_bus().register_default_events()
    instruct = GetExtraInformation(content="Tell me the use case in 1 sentence.", summary="Requesting details about the use case")
    responses = await user.handle(Message(role="Bot", content=instruct.json(), data=instruct))
    assert isinstance(responses[0].data, SearchInternetQuery)


@pytest.mark.asyncio
async def test_respond_user_str():
    async def respond_to_user(query: str, role: Role) -> str:
        """Respond to user."""
        response = await role.aask(query, str)
        return response
        
    role = Role(name="Alice",
                profile="Customer service",
                goal="Efficiently communicate with the user and translate the user's needs to technical requirements",
                constraints=None,
                actions=[respond_to_user])
    responses = await role.handle("Say hello")
    assert responses


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Time    : 2023/6/1 12:41
@Author  : alexanderwu
@File    : logs.py
"""

import sys

from loguru import logger as _logger

from schema_agents.const import PROJECT_ROOT


def define_log_level(print_level="INFO", logfile_level="DEBUG"):
    """Define log level."""
    _logger.remove()
    _logger.add(sys.stderr, level=print_level)
    _logger.add(PROJECT_ROOT / 'logs/log.txt', level=logfile_level)
    return _logger


logger = define_log_level()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Time    : 2023/4/30 20:57
@Author  : alexanderwu
@File    : __init__.py
"""

from schema_agents.memory.memory import Memory
from schema_agents.memory.long_term_memory import LongTermMemory
from schema_agents.memory.faiss_store import FaissStore




import ray
import asyncio
import argparse
from imjoy_rpc.hypha import connect_to_server
import cloudpickle

async def register_ray(server, ray_address=None):
    if ray_address:
        ray.init(address=ray_address)

    print(server.config)
    server.register_codec({
      "name": "RayObjectRef",
      "type": ray.ObjectRef,
      "encoder": lambda x: {"id": x.hex()},
      "decoder": lambda x: ray.ObjectRef(x["id"]),
    })

    def encode_ray_function(x):
        pass

    def decode_ray_function(x):
        return ray.remote(x)

    server.register_codec({
        "name": "RayRemoteFunction",
        "type": ray.remote_function.RemoteFunction,
        "encoder": encode_ray_function,
        "decoder": decode_ray_function,
    })

    async def remote(f):
        func = cloudpickle.loads(f)
        return ray.remote(func)

    await server.register_service(
        {
            "name": "Ray",
            "id": "ray",
            "config": {
                "visibility": "public",
                "require_context": False,
                "run_in_executor": True,
            },
            "put": lambda x: ray.put(x),
            "get": lambda x: ray.get(x),
            "remote": remote,
        }
    )
    print("Ray service registered.")
    
async def test_ray(server):
    svc = await server.get_service("ray")
    f = cloudpickle.dumps(lambda x: x + 1)
    func = await svc.remote(f)
    print(await func(1))
    print("Ray service is working.")

async def start_function_launcher(server_url, workspace, token, ray_address):
    server = await connect_to_server(
        {
            "name": "function client",
            "server_url": server_url,
            "token": token,
            "workspace": workspace,
        }
    )
    
    print("Registering Ray service...")
    await register_ray(server, ray_address=ray_address)
    
    await test_ray(server)
    
    print("Ray service is ready.")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--server-url", default="https://ai.imjoy.io")
    parser.add_argument("--workspace", default=None)
    parser.add_argument("--token", default=None)
    parser.add_argument("--ray-server", default="auto")
    opts = parser.parse_args()

    loop = asyncio.get_event_loop()
    task = loop.create_task(
        start_function_launcher(
            ray_address=opts.ray_server,
            server_url=opts.server_url,
            token=opts.token,
            workspace=opts.workspace,
        )
    )
    loop.run_forever()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Time    : 2023/5/12 00:30
@Author  : alexanderwu
@File    : software_team.py
"""
from typing import Union, Optional
from pydantic import BaseModel, Field 
from schema_agents.config import CONFIG
from schema_agents.logs import logger
from schema_agents.role import Role
from schema_agents.schema import Message
from schema_agents.utils.common import NoMoneyException, EventBus
import uuid

class Team(Role):
    """
    Team: Possesses a team, SOP (Standard Operating Procedures), and a platform for instant messaging,
    dedicated to writing executable code.
    """
    def __init__(self, *args, investment=10, **kwargs):
        super().__init__(*args, **kwargs)
        assert not kwargs.get("actions"), "Team can't have actions."
        self._investment: float = investment
        self._roles = []

    def hire(self, roles: list[Role]):
        """Hire roles to cooperate"""
        roles = list(set(roles))
        for role in roles:
            role.set_event_bus(self._event_bus)
        self._roles.extend(roles)
    
    def fire(self, role: Role):
        """Fire a role"""
        self._roles.remove(role)
        role.set_event_bus(None)

    def invest(self, investment: float):
        """Invest team. raise NoMoneyException when exceed max_budget."""
        self._investment += investment
        CONFIG.max_budget = self._investment
        logger.info(f'Investment: ${self._investment}.')

    def _check_balance(self):
        if CONFIG.total_cost > CONFIG.max_budget:
            raise NoMoneyException(CONFIG.total_cost, f'Insufficient funds: {CONFIG.max_budget}')

    def _save(self):
        logger.info(self.json())
    
    def can_handle(self, message: Message) -> bool:
        """Check if the team can handle the message."""
        if not self._roles:
            raise ValueError(f"Team {self.name} haven't hired anyone.")
        for role in self._roles:
            if role.can_handle(message):
                return True
        return False

    async def handle(self, message: Union[str, Message]) -> list[Message]:
        """Handle message"""
        if isinstance(message, str):
            message = Message(role="User", content=message)
        if not self.can_handle(message):
            raise ValueError(f"Team {self.name} can't handle message: {message}")
        _session_id = message.session_id or str(uuid.uuid4())
        message.session_history.append(_session_id)
        messages = []
        def on_message(new_msg):
            self._check_balance()
            if _session_id in new_msg.session_history:
                messages.append(new_msg)

        self._event_bus.on("message", on_message)
        try:
            # start to process the message
            await self._event_bus.aemit("message", message)
        finally:
            self._event_bus.off("message", on_message)
        return messages
        


if __name__ == "__main__":
    import asyncio
    team = Team(name="Software Team")
    
    async def test(message: str):
        """Test action"""
        print(message)

    team.hire([Role(profile="QA", actions=[test])])
    asyncio.run(team.handle(Message(role="User", content="I want to build a software team.")))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pytest
from schema_agents.role import Role
from schema_agents.schema import Message
from pydantic import BaseModel, Field
from typing import List


class Recipe(BaseModel):
    """A recipe."""
    name: str = Field(description="The name of the recipe.")
    ingredients: List[str] = Field(description="The list of ingredients.")
    instructions: str = Field(description="The instructions for making the recipe.")
    rating: float = Field(description="The rating of the recipe.")


class CookBook(BaseModel):
    """Creating a recipe book with a list of recipes based on the user's query."""
    name: str = Field(description="The name of the recipe book.")
    recipes: List[Recipe] = Field(description="The list of recipes in the book.")


async def respond_to_user(query: str, role: Role = None) -> CookBook:
    """Respond to user's request by recipe book."""
    response = await role.aask([{"role": "user", "content": [
        {"type": "text", "text": "Cook something like this"},
        {
            "type": "image_url",
            "image_url": {
            "url": "https://upload.wikimedia.org/wikipedia/commons/thumb/d/dd/Gfp-wisconsin-madison-the-nature-boardwalk.jpg/2560px-Gfp-wisconsin-madison-the-nature-boardwalk.jpg",
            },
        },
    ]}], str)
    return response
    
@pytest.mark.asyncio
async def test_vision():
    alice = Role(
        name="Alice",
        profile="Cooker",
        goal="Your goal is to listen to user's request and propose recipes for making the most delicious meal for thanksgiving.",
        constraints=None,
        actions=[respond_to_user],
        model="gpt-4-vision-preview",
    )
    event_bus = alice.get_event_bus()
    event_bus.register_default_events()
    responses = await alice.handle(Message(content="make something to surprise our guest from Stockholm.", role="User"))
    print(responses)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations

import asyncio
import inspect
import json
import uuid
import traceback
import types
import typing
from functools import partial
from inspect import signature
from typing import Iterable, Optional, Union, Callable
from pydantic import BaseModel, create_model, Field, ValidationError

from schema_agents.logs import logger
from schema_agents.utils import parse_special_json, schema_to_function
from schema_agents.llm import LLM
from schema_agents.schema import Message, RoleSetting, Session
from schema_agents.memory.long_term_memory import LongTermMemory
from schema_agents.utils import dict_to_pydantic_model
from schema_agents.utils.common import EventBus
from schema_agents.utils.common import current_session
from contextlib import asynccontextmanager
from contextvars import copy_context

PREFIX_TEMPLATE = """You are a {profile}, named {name}, your goal is {goal}, and the constraint is {constraints}. """


@asynccontextmanager
async def create_session_context(id=None, role_setting=None):
    pre_session = current_session.get()
    if pre_session:
        id = id or pre_session.id
        role_setting = role_setting or pre_session.role_setting
    current_session.set(Session(id=id, role_setting=role_setting))
    yield copy_context()
    current_session.set(pre_session)



class Role:
    """Role is a person or group who has a specific job or purpose within an organization."""

    def __init__(
        self,
        name="",
        profile="",
        goal="",
        constraints=None,
        desc="",
        icon="🤖",
        long_term_memory: Optional[LongTermMemory] = None,
        event_bus: EventBus = None,
        actions: list[Callable] = None,
        **kwargs,
    ):
        self._llm = LLM(**kwargs)
        self._setting = RoleSetting(
            name=name,
            profile=profile,
            goal=goal,
            constraints=constraints,
            desc=desc,
            icon=icon,
        )
        self._states = []
        self._actions = actions or []
        self._role_id = str(self._setting)
        self._input_schemas = []
        self._output_schemas = []
        self._action_index = {}
        self._user_support_actions = []
        self._watch_schemas = set()
        self.long_term_memory = long_term_memory
        if event_bus:
            self.set_event_bus(event_bus)
        else:
            self.set_event_bus(
                EventBus(f"{self._setting.profile} - {self._setting.name}")
            )
        self._init_actions(self._actions)

    @property
    def llm(self):
        return self._llm

    def _reset(self):
        self._states = []
        self._actions = []

    def _watch(self, schemas: Iterable[Union[str, BaseModel]]):
        """Watch actions."""
        self._watch_schemas.update(schemas)

    def set_event_bus(self, event_bus: EventBus):
        """Set event bus."""
        self._event_bus = event_bus

        async def handle_message(msg):
            if msg.data and type(msg.data) in self._watch_schemas:
                if msg.cause_by not in self._action_index[type(msg.data)]:
                    await self.handle(msg)
            elif msg.data is None and str in self._watch_schemas:
                if msg.cause_by not in self._action_index[str]:
                    await self.handle(msg)

        self._event_bus.on("message", handle_message)
        logger.info(f"Mounting {self._setting} to event bus: {self._event_bus.name}.")

    def get_event_bus(self):
        """Get event bus."""
        return self._event_bus

    @property
    def profile(self):
        """Get profile."""
        return self._setting.profile

    def _get_prefix(self):
        """Get prefix."""
        if self._setting.desc:
            return self._setting.desc
        return PREFIX_TEMPLATE.format(**self._setting.dict())

    @property
    def user_support_actions(self):
        return self._user_support_actions

    @property
    def prefix(self):
        return self._get_prefix()

    def _extract_schemas(self, function):
        sig = signature(function)
        positional_annotation = [
            p.annotation
            for p in sig.parameters.values()
            if p.kind == p.POSITIONAL_OR_KEYWORD
        ][0]
        assert (
            positional_annotation == str
            or isinstance(positional_annotation, typing._UnionGenericAlias)
            or issubclass(positional_annotation, BaseModel)
        ), f"Action only support pydantic BaseModel, typing.Union or str, but got {positional_annotation}"
        output_schemas = (
            [sig.return_annotation]
            if not isinstance(sig.return_annotation, typing._UnionGenericAlias)
            else list(sig.return_annotation.__args__)
        )
        input_schemas = (
            [positional_annotation]
            if not isinstance(positional_annotation, typing._UnionGenericAlias)
            else list(positional_annotation.__args__)
        )
        return input_schemas, output_schemas

    def _init_actions(self, actions):
        self._output_schemas = []
        self._input_schemas = []
        for action in actions:
            if isinstance(action, partial):
                action.__doc__ = action.func.__doc__
                action.__name__ = action.func.__name__
            assert action.__doc__, "Action must have docstring"
            assert isinstance(
                action, (partial, types.FunctionType, types.MethodType)
            ), f"Action must be function, but got {action}"
            input_schemas, output_schemas = self._extract_schemas(action)
            self._output_schemas += output_schemas
            self._input_schemas += input_schemas
            for schema in input_schemas:
                if schema not in self._action_index:
                    self._action_index[schema] = [action]
                else:
                    self._action_index[schema].append(action)
            # mark as user support action if the input schema is str
            if str in input_schemas:
                self._user_support_actions.append(action)
        self._output_schemas = list(set(self._output_schemas))
        self._input_schemas = list(set(self._input_schemas))
        self._watch(self._input_schemas)

        self._reset()
        for idx, action in enumerate(actions):
            self._actions.append(action)
            self._states.append(f"{idx}. {action}")

    async def _run_action(self, action, msg):
        sig = signature(action)
        keys = [
            p.name
            for p in sig.parameters.values()
            if p.kind == p.POSITIONAL_OR_KEYWORD and p.annotation == Role
        ]
        kwargs = {k: self for k in keys}
        pos = [
            p
            for p in sig.parameters.values()
            if p.kind == p.POSITIONAL_OR_KEYWORD and p.annotation != Role
        ]
        for p in pos:
            if not msg.data and isinstance(msg.content, str):
                kwargs[p.name] = msg.content
                msg.processed_by.add(self)
                break
            elif msg.data and isinstance(
                msg.data,
                p.annotation.__args__
                if isinstance(p.annotation, typing._UnionGenericAlias)
                else p.annotation,
            ):
                kwargs[p.name] = msg.data
                msg.processed_by.add(self)
                break
            if p.name not in kwargs:
                kwargs[p.name] = None

        if inspect.iscoroutinefunction(action):
            return await action(**kwargs)
        else:
            return action(**kwargs)

    def can_handle(self, message: Message) -> bool:
        """Check if the role can handle the message."""
        context_class = (
            message.data.__class__ if message.data else type(message.content)
        )
        if context_class in self._input_schemas:
            return True
        return False

    # @retry(stop=stop_after_attempt(2), wait=wait_fixed(1))
    async def handle(self, msg: Union[str, Message]) -> list[Message]:
        """Handle message"""
        if isinstance(msg, str):
            msg = Message(role="User", content=msg)
        if not self.can_handle(msg):
            raise ValueError(
                f"Invalid message, the role {self._setting} cannot handle the message: {msg}"
            )
        _session_id = msg.session_id or str(uuid.uuid4())
        msg.session_history.append(_session_id)
        messages = []

        def on_message(new_msg):
            if _session_id in new_msg.session_history:
                messages.append(new_msg)

        self._event_bus.on("message", on_message)
        try:
            context_class = msg.data.__class__ if msg.data else type(msg.content)
            responses = []
            if context_class in self._input_schemas:
                actions = self._action_index[context_class]
                for action in actions:
                    responses.append(self._run_action(action, msg))
            async with create_session_context(
                id=msg.session_id, role_setting=self._setting
            ):
                responses = await asyncio.gather(*responses)

            outputs = []
            for response in responses:
                if not response:
                    continue
                # logger.info(response)
                if isinstance(response, str):
                    output = Message(
                        content=response,
                        role=self.profile,
                        cause_by=action,
                        session_history=msg.session_history.copy(),
                    )
                else:
                    assert isinstance(
                        response, BaseModel
                    ), f"Action must return str or pydantic BaseModel, but got {response}"
                    output = Message(
                        content=response.json(),
                        data=response,
                        session_history=msg.session_history.copy(),
                        role=self.profile,
                        cause_by=action,
                    )
                # self._rc.memory.add(output)
                # logger.debug(f"{response}")
                outputs.append(output)
                async with create_session_context(
                    id=msg.session_id, role_setting=self._setting
                ):
                    await self._event_bus.aemit("message", output)
        finally:
            self._event_bus.off("message", on_message)
        return messages

    def _normalize_messages(self, req):
        input_schema = []
        if isinstance(req, str):
            messages = [{"role": "user", "content": req}]
        elif isinstance(req, dict):
            messages = [req]
        elif isinstance(req, BaseModel):
            input_schema.append(req.__class__)
            messages = [
                {
                    "role": "function",
                    "name": req.__class__.__name__,
                    "content": req.json(),
                }
            ]
        else:
            assert isinstance(req, list)
            messages = []
            for r in req:
                if isinstance(r, str):
                    messages.append({"role": "user", "content": r})
                elif isinstance(r, dict):
                    messages.append(r)
                elif isinstance(r, BaseModel):
                    input_schema.append(r.__class__)
                    messages.append(
                        {
                            "role": "function",
                            "name": r.__class__.__name__,
                            "content": r.json(),
                        }
                    )
                else:
                    raise ValueError(f"Invalid request {r}")
        return messages, input_schema

    def _parse_outputs(self, response, output_types, parallel_call):
        if response["type"] == "function_call":
            func_call = response["function_call"]
            assert func_call["name"] in [
                s.__name__ for s in output_types
            ], f"Invalid function name: {func_call['name']}"
            idx = [s.__name__ for s in output_types].index(func_call["name"])
            arguments = parse_special_json(func_call["arguments"])
            function_args = output_types[idx].parse_obj(arguments)
            return function_args, {
                "system_fingerprint": response["system_fingerprint"],
                "function_call": func_call,
            }
        elif response["type"] == "tool_calls":
            tool_calls = response["tool_calls"]
            functions = []
            ids = []
            for tool_call in tool_calls:
                assert tool_call["type"] == "function"
                func_call = tool_call["function"]
                assert func_call["name"] in [
                    s.__name__ for s in output_types
                ], f"Invalid function name: {func_call['name']}"
                idx = [s.__name__ for s in output_types].index(func_call["name"])
                arguments = parse_special_json(func_call["arguments"])
                if len(arguments.keys()) == 1 and "_" in arguments.keys():
                    arguments = arguments["_"]
                model = output_types[idx]
                try:
                    fargs = model.parse_obj(arguments)
                except ValidationError:
                    model_schema = model.schema()
                    keys = model_schema["properties"].keys()
                    # only one argument
                    if len(keys) == 1:
                        fargs = model.parse_obj(
                            {list(keys)[0]: arguments}
                        )
                    else:
                        raise
                functions.append(fargs)
                ids.append(tool_call["id"])
            if not parallel_call:
                functions = functions[0]
            return functions, {
                "tool_ids": ids,
                "system_fingerprint": response["system_fingerprint"],
                "tool_calls": tool_calls,
            }

    async def acall(self, req, tools, output_schema=None):
        tool_inputs_models = []
        arg_names = []
        messages, _ = self._normalize_messages(req)
        tool_output_models = []
        for tool in tools:
            assert callable(tool), "Tools must be callable functions"
            sig = signature(tool)
            var_positional = [
                p.name for p in sig.parameters.values() if p.kind == p.VAR_POSITIONAL
            ]
            kwargs_args = [
                p.name for p in sig.parameters.values() if p.kind != p.VAR_POSITIONAL
            ]
            arg_names.append((var_positional, kwargs_args))
            names = [p.name for p in sig.parameters.values()]
            types = [sig.parameters[name].annotation for name in names]
            defaults = []
            for i, name in enumerate(names):
                if sig.parameters[name].default == inspect._empty:
                    defaults.append(Field(..., description=types[i].__doc__))
                else:
                    defaults.append(
                        Field(
                            sig.parameters[name].default, description=types[i].__doc__
                        )
                    )
            tool_output_models.append(sig.return_annotation)
            tool_inputs_models.append(
                dict_to_pydantic_model(
                    tool.__name__,
                    {names[i]: (types[i], defaults[i]) for i in range(len(names))},
                    tool.__doc__,
                )
            )

        tool_calls, metadata = await self.aask(
            messages, tool_inputs_models, use_tool_calls=True, return_metadata=True
        )
        tool_ids = metadata["tool_ids"]
        messages.append({"role": "assistant", "tool_calls": metadata["tool_calls"]})
        promises = []
        for fargs in tool_calls:
            idx = tool_inputs_models.index(fargs.__class__)
            args_ns, kwargs_ns = arg_names[idx]
            args = [getattr(fargs, name) for name in args_ns]
            kwargs = {name: getattr(fargs, name) for name in kwargs_ns}
            tool = tools[idx]
            if inspect.iscoroutinefunction(tool):
                promises.append(tool(*args, **kwargs))
            else:
                loop = asyncio.get_running_loop()

                def wrap_func():
                    return tool(*args, **kwargs)

                promises.append(loop.run_in_executor(None, wrap_func))
        results = await asyncio.gather(*promises)

        extra_schemas = []
        result_dict = {}
        for call_id, result in enumerate(results):
            fargs = tool_calls[call_id]
            idx = tool_inputs_models.index(fargs.__class__)
            result_dict[tools[idx]] = result
            if isinstance(result, BaseModel):
                extra_schemas.append(result.__class__)
            messages.append(
                {
                    "tool_call_id": tool_ids[call_id],
                    "role": "tool",
                    "name": tools[idx].__name__,
                    "content": result.json()
                    if isinstance(result, BaseModel)
                    else str(result),
                }
            )  # extend conversation with function response

        if not output_schema:
            return result_dict

        return await self.aask(
            messages,
            output_schema=output_schema,
            use_tool_calls=True,
            extra_schemas=extra_schemas,
        )

    # @retry(stop=stop_after_attempt(2), wait=wait_fixed(1))
    async def aask(
        self,
        req,
        output_schema=None,
        prompt=None,
        use_tool_calls=False,
        return_metadata=False,
        extra_schemas=None,
    ):
        
        assert extra_schemas is None or isinstance(extra_schemas, list)
        output_schema = output_schema or str
        messages, input_schema = self._normalize_messages(req)
        if use_tool_calls:
            assert (
                output_schema is str
                or isinstance(output_schema, (tuple, list))
                or isinstance(output_schema, typing._UnionGenericAlias)
                or issubclass(output_schema, BaseModel)
            )
        else:
            assert (
                output_schema is str
                or isinstance(output_schema, typing._UnionGenericAlias)
                or issubclass(output_schema, BaseModel)
            )

        if extra_schemas:
            input_schema += extra_schemas

        if input_schema:
            sch = ",".join([f"`{i.__name__}`" for i in input_schema])
            prefix = f"Please generate a response based on results from: {sch}. "
        else:
            prefix = ""

        parallel_call = False
        if output_schema is str:
            output_types = []
            prompt = prompt or f"{prefix}"
            if prompt:
                messages.append({"role": "user", "content": f"{prompt}"})
        elif isinstance(output_schema, (tuple, list)):
            # A list means can call multiple functions in parallel
            assert (
                use_tool_calls
            ), "Please set `use_tool_calls` to True when passing a list of output schemas."
            output_types = list(output_schema)
            parallel_call = True
            schema_names = ",".join([f"`{s.__name__}`" for s in output_types])
            avoid_schema_names = ",".join([f"`{s.__name__}`" for s in input_schema])
            if avoid_schema_names:
                prompt = prompt or f"{prefix}You MUST call one or more of these functions: {schema_names}. DO NOT return text directly or call any of the following functions: {avoid_schema_names}."
            else:
                prompt = prompt or f"{prefix}You MUST call one or more of these functions: {schema_names}. DO NOT return text directly."

            messages.append({"role": "user", "content": f"{prompt}"})
        elif isinstance(output_schema, typing._UnionGenericAlias):
            # A union type means can only call one function at a time
            output_types = list(output_schema.__args__)
            schema_names = ",".join([f"`{s.__name__}`" for s in output_types])
            avoid_schema_names = ",".join([f"`{s.__name__}`" for s in input_schema])
            if avoid_schema_names:
                prompt = prompt or f"{prefix}You MUST call one of these functions: {schema_names}. DO NOT return text directly or call any of the following functions: {avoid_schema_names}."
            else:
                prompt = prompt or f"{prefix}You MUST call one of these functions: {schema_names}. DO NOT return text directly."
            messages.append({"role": "user", "content": f"{prompt}"})
        else:
            output_types = [output_schema]
            prompt = (
                prompt
                or f"{prefix}You MUST call the `{output_schema.__name__}` function."
            )
            messages.append({"role": "user", "content": f"{prompt}"})
        system_msgs = [self._get_prefix()]

        if output_schema is str:
            function_call = "none"
            response = await self._llm.aask(
                messages,
                system_msgs,
                functions=[schema_to_function(s) for s in input_schema],
                function_call=function_call,
                event_bus=self._event_bus,
            )
            assert isinstance(
                response, str
            ), f"Invalid response type, it must be a string"
            return response
        functions = [schema_to_function(s) for s in set(output_types + input_schema)]
        schema_names = ",".join([f"`{s.__name__}`" for s in output_types])
        avoid_schema_names = ",".join([f"`{s.__name__}`" for s in input_schema])
        if len(output_types) == 1:
            function_call = {"name": output_types[0].__name__}
        else:
            function_call = "auto"
        response = await self._llm.aask(
            messages,
            system_msgs,
            functions=functions,
            function_call=function_call,
            event_bus=self._event_bus,
            use_tool_calls=use_tool_calls,
        )

        try:
            assert not isinstance(
                response, str
            ), f"Invalid response. {prompt}"
            function_args, metadata = self._parse_outputs(
                response, output_types, parallel_call
            )
            if return_metadata:
                return function_args, metadata
            return function_args
        except Exception:
            logger.error(
                f"Failed to parse the response, error:\n{traceback.format_exc()}\nPlease regenerate to fix the error."
            )
            messages.append({"role": "assistant", "content": str(response)})
            messages.append(
                {
                    "role": "user",
                    "content": f"Failed to parse the response, error:\n{traceback.format_exc()}\nPlease regenerate to fix the error.",
                }
            )
            response = await self._llm.aask(
                messages,
                system_msgs,
                functions=functions,
                function_call=function_call,
                event_bus=self._event_bus,
                use_tool_calls=use_tool_calls,
            )
            function_args, metadata = self._parse_outputs(
                response, output_types, parallel_call
            )
            if return_metadata:
                return function_args, metadata
            return function_args



import os
# from simpleaichat import AIChat
import re
import inspect
# To remove jupyter warning.
os.environ["JUPYTER_PLATFORM_DIRS"] = "1"
from jupyter_client.manager import start_new_kernel
from jupyter_client.utils import run_sync
from .msgspec_v5 import validate_message
import os
import re
import shortuuid
import base64
import json
import logging

logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger(__name__)

def strip_ansi_codes(s):
    return re.sub(r'\x1B[@-_][0-?]*[ -/]*[@-~]', '', s)

TIMEOUT=15

INIT_CODE = """%matplotlib inline
%cd {work_dir}
import sys
sys.path.append('{work_dir}')
"""

KERNEL_INFO_CODE="""
import platform
import pkg_resources
python_version = platform.python_version()
# Get list of installed packages and their versions
installed_packages = [f'{d.project_name}-{d.version}' for d in pkg_resources.working_set]
print(f"PYTHON VERSION: {python_version}, INSTALLED PACKAGES: {','.join(installed_packages)}")
"""

SYSTEM_PROMPT = """
Act as a professional python programmer, you help user to generate useful python code.
When you send a message containing Python code (quoted with ```python) to the user, it will be executed in a stateful Jupyter notebook environment. Jupyter will respond with the output of the execution or time out after 120.0 
seconds. The drive at '{work_dir}' can be used to save and persist user files. Internet access for this session is disabled. Do not make external web requests or API calls as they will fail.
Input should be valid python code for this particular jupyter notebook kernel. Here is more details about the jupyter environment:
{kernel_info}

You MUST ensure that the code generated is compatible with the Jupyter Notebook environment.
"""

FORMAT_INSTRUCTION = """If you want to see the output of a value, you should print it out with `print(...)`.
Example:
User: print hello world
```
print("hello world")
```
In the following message, the user might provide feedback on the result of the code execution via `Jupyter: ....`. The feedback will be used to improve the model.
"""

ERROR_FIX_PROMPT = "Error while executing the code above: {error} \nPlease fix the error and try again. If needed, you can use `import subprocess;subprocess.check_call([sys.executable, '-m', 'pip', 'install', 'new_package'])` to install a new package; DO NOT generate the same code or error twice."

def ensure_sync(func):
    if inspect.iscoroutinefunction(func):
        return run_sync(func)
    return func

def extract_code_blocks(text):
    pattern = r"```(?:python)?\s*([\s\S]*?)\s*```"
    matches = re.findall(pattern, text, re.MULTILINE)
    return matches

def extract_stdout(results):
    output = results["outputs"]
    return "\n".join([item['stdout'] for item in output if 'stdout' in item])

def extract_stderr(results):
    output = results["outputs"]
    return "\n".join([item['stderr'] for item in output if 'stderr' in item])

def extract_error(results):
    output = results["outputs"]
    return "\n".join([item['error'] for item in output if 'error' in item])

def extract_execute_result(results):
    output = results["outputs"]
    return "\n".join([item['execute_result'] for item in output if 'execute_result' in item])

def extract_display_data(results):
    output = results["outputs"]
    return "\n".join([item['display_data'] for item in output if 'display_data' in item])

class CodeInterpreter:
    MAIN_SESSION_ID = "main"

    def __init__(self, bot=None, session_id=None, work_dir_root="./.code-interpreter"):
        self.store = {}
        self._last_output = ""
        self.session_id = shortuuid.uuid() if session_id is None else session_id
        self.work_dir = os.path.join(work_dir_root, self.session_id)
        assert os.path.exists(work_dir_root), f"work_dir_root {work_dir_root} does not exist"
        os.makedirs(self.work_dir, exist_ok=True)
        self.bot = bot
        self.initialize()

    def initialize(self):
        self.km, self.kc = start_new_kernel(kernel_name="python")
        results = self.execute_code(INIT_CODE.format(work_dir=self.work_dir) + KERNEL_INFO_CODE)
        assert results['status'] == "ok", f"failed to get kernel info, output: {results}"
        kernel_info = extract_stdout(results).replace('\n', ' ')
        self.system_prompt = SYSTEM_PROMPT.format(work_dir=self.work_dir, kernel_info=kernel_info)

    def reset(self):
        self.tearDown()
        self.initialize()
        # self.bot.reset_session()

    def tearDown(self):
        # Check if the kernel client and kernel manager exist
        if self.kc:
            # Send a shutdown request to the kernel
            self.kc.shutdown()
            self.kc = None

        if self.km:
            # Shut down the kernel
            self.km.shutdown_kernel()
            self.km = None

    def assertEqual(self, a, b):
        assert a == b, f"{a} != {b}"

    def get_non_kernel_info_reply(self, timeout=None):
        while True:
            reply = self.kc.get_shell_msg(timeout=timeout)
            if reply["header"]["msg_type"] != "kernel_info_reply":
                return reply

    def get_data_from_message(self, output_msgs, msg_type):
        data_list = []
        for msg in output_msgs:
            if msg["msg_type"] == msg_type:
                data_list.append(msg["content"]["data"])
            else:
                continue
        if data_list:
            return data_list
        
    def execute_code(self, code, timeout=TIMEOUT, silent=False, store_history=True, stop_on_error=True):
        """Execute python code in a jupyter kernel and return the outcome.
        Arguments:
         - code (str): The code to execute, need to be a valid python code.
        Returns: a dict containing the STDOUT, STDERR, and the execution status.
        """
        msg_id = self.kc.execute(
            code=code, silent=silent, store_history=store_history, stop_on_error=stop_on_error
        )

        try:
            reply = self.get_non_kernel_info_reply(timeout=timeout)
            validate_message(reply, "execute_reply", msg_id)
        except Exception:
            self.reset()
            raise RuntimeError(f"The Jupyter kernel may have stalled or crashed, failing to provide an execute_reply within {timeout} seconds. It has been forcibly restarted! Please avoid using functions that interact with desktop UI, such as cv2.imshow, which could potentially crash the kernel.")

        busy_msg = ensure_sync(self.kc.iopub_channel.get_msg)(timeout=1)
        validate_message(busy_msg, "status", msg_id)
        self.assertEqual(busy_msg["content"]["execution_state"], "busy")

        output_msgs = []
        while True:
            msg = ensure_sync(self.kc.iopub_channel.get_msg)(timeout=0.1)
            validate_message(msg, msg["msg_type"], msg_id)
            if msg["msg_type"] == "status":
                self.assertEqual(msg["content"]["execution_state"], "idle")
                break
            elif msg["msg_type"] == "execute_input":
                self.assertEqual(msg["content"]["code"], code)
                continue
            output_msgs.append(msg)
        # Process output messages
        tb = "\n".join(reply["content"].get("traceback", []))
        if tb:
            tb = strip_ansi_codes(tb)
        results = {"status": reply["content"]["status"], "traceback": tb, "outputs": []}
        for msg in output_msgs:
            if msg["msg_type"] == "stream":
                if msg["content"]["name"] == "stdout":
                    results["outputs"].append({"stdout": msg["content"]["text"]})
                elif msg["content"]["name"] == "stderr":
                    results["outputs"].append({"stderr": msg["content"]["text"]})
            elif msg["msg_type"] == "error":
                results["outputs"].append({"error": f'{msg["content"]["ename"]}: {msg["content"]["evalue"]}'})
            elif "data" in msg["content"]:
                results["outputs"].append({msg["msg_type"]: msg["content"]["data"]})
            else:
                logger.warning("Unknown message type: %s", msg["msg_type"])

        # Here we process the execution results
        for output in results["outputs"]:
            if output.get("display_data"):
                display_data = output.get("display_data")
                display_type = "display_data"
            elif output.get("execute_result"):
                display_data = output.get("execute_result")
                display_type = "execute_result"
            else:
                continue
            key = shortuuid.uuid() # Define a unique key for the display_data
            for type_name in list(display_data.keys()):
                data = display_data[type_name]
                if type_name == 'text/plain':
                    if len(data) > 1000:
                        # truncate the text if it is too long
                        file_name = f"{key}.txt"
                        with open(os.path.join(self.work_dir, file_name), 'w') as file:
                            file.write(data)
                        display_data[type_name] = data[:1000] + "..."

                elif type_name == 'image/png':
                    file_extension = 'png'
                    image_bytes = base64.b64decode(data)
                    file_name = f"{key}.{file_extension}"
                    with open(os.path.join(self.work_dir, file_name), 'wb') as file:
                        file.write(image_bytes)
                    del display_data[type_name]
                elif type_name == 'image/jpeg' or type_name == 'image/jpg':
                    file_extension = 'jpeg'
                    image_bytes = base64.b64decode(data)
                    file_name = f"{key}.{file_extension}"
                    with open(os.path.join(self.work_dir, file_name), 'wb') as file:
                        file.write(image_bytes)
                    del display_data[type_name]
                else:
                    # for other types of display_data, save them into a JSON file
                    file_extension = 'json'
                    file_name = f"{key}.{file_extension}"
                    with open(os.path.join(self.work_dir, file_name), 'w') as file:
                        json.dump({'type': type_name, 'data': data}, file)

        return results
        
    def get_store(self, key):
        return self.store[key]

    def fork_session(self, session_id):
        sess = self.bot.get_session(self.MAIN_SESSION_ID)
        sess_dict = sess.model_dump(
            exclude={"auth", "api_url", "input_fields"},
            exclude_none=True,
        )
        sess_dict["id"] = session_id
        self.bot.new_session(**sess_dict)
        last_message_idx = len(sess.messages) - 1
        return last_message_idx 
    
    def merge_session(self, target_session_id, messages, last_output):
        sess = self.bot.get_session(target_session_id)
        sess.messages.extend(messages)
        # save the last output so we can prepend it to the next user message
        self._last_output = f"Jupyter: {last_output}\n"
        
    def save_session(self, path):
        # TODO: save the last output
        self.bot.save_session(path)

    def _create_summary(self, result):
        """
        Create a text summary to describe the execution result.
        """
        return json.dumps({k: result[k] for k in result.keys() if result[k]}, indent=1)

    def execute_query(self, query: str) -> str:
        """Generate python code based on a input query string and execute it in a jupyter kernel. The output of the execution containing `stdout`, `stderr`, `display_data` will be returned as a string"""
        logger.info("Executing query: %s", query)
        self.fork_session("dev")
        response = self.bot(f"{self._last_output}User: {query}", id="dev")
        query_message = self.bot.get_session("dev").messages[-2]
        max_iteration = 5
        while True:
            logger.info("Executing python code:\n---\n%s\n---\n", response)
            try:
                command = "\n".join(extract_code_blocks(response))
                if not command:
                    return response
                assert command, "no python code is generated (Note: remember to quote it as a markdown code block)"
                exec_result = self.execute_code(command)
                
                if exec_result["status"] == "ok":
                    exec_result["status"] = "successfully executed"
                    exec_result["code"] = command
                    output = self._create_summary(exec_result)
                    response_message = self.bot.get_session("dev").messages[-1]
                    # format the response message
                    response_message.content = f"```python\n{command}```\n{output}"
                    # merge the two correct messages into the main session
                    self.merge_session(self.MAIN_SESSION_ID, [query_message, response_message], output)
                    output = f"Jupyter: {output}"
                    break
                else:
                    exec_result["status"] = "failed to execute the code."
                    output = self._create_summary(exec_result)
                    output = f"Jupyter: {output}"
                    raise RuntimeError(output)
            except RuntimeError:
                max_iteration -= 1
                # output = traceback.format_exc()
                logger.info("RuntimeError: %s", output)
                if max_iteration <= 0:
                    break
                response = self.bot(ERROR_FIX_PROMPT.format(error=output), id="dev")
        if max_iteration <= 0 and output.startswith("Error:"):
            output += f"After trying {max_iteration} times, the tool still failed to produce code for the task."
            logger.error(output)
        return output

def create_mock_client(work_dir_root = "./.data", form_data=None):
    os.makedirs(work_dir_root, exist_ok=True)
    ex = CodeInterpreter(work_dir_root=work_dir_root)
    
    if not form_data:
        form_data = {}
    
    class MockDialog():
        def __init__(self, response) -> None:
            self.response = response

        async def get_data(self):
            return self.response
    class MockClient():
        def __init__(self) -> None:
            self.system_prompt =  ex.system_prompt

        async def newMessage(self, data):
            print(data)

        async def initialize(self, data):
            pass
        
        async def showMessage(self, message):
            print(message)
        
        async def appendText(self, data):
            print(data)

        async def executeScript(self, data):
            script = data['script']
            del data['script']
            return ex.execute_code(script, **data)

        async def showDialog(self, src, config=None, data=None):
            print(src, config, data)
            return MockDialog({"formData": form_data})

        async def createWindow(self, src, config=None, data=None):
            print(src, config, data)
            return MockDialog({})

    return MockClient()
    
if __name__ == "__main__":
    work_dir_root = "./.data"
    os.makedirs(work_dir_root, exist_ok=True)
    ex = CodeInterpreter(work_dir_root=work_dir_root)
    try:
        # result = ex.execute_code("print('hello world')")
        # print("result for print hello world:", result)
        # result = ex.execute_code("import numpy as np; print(np.random.rand(3))")
        # print("result for numpy array", result)
        # result = ex.execute_code("import matplotlib.pyplot as plt; plt.plot([1,2,3,4]); plt.show()")
        # print("result for plotting", result)
        result = ex.execute_query("make a simple plot")
        print("simple plot:", result)
        result = ex.execute_query("save the plot into a png file")
        print("save plot:", result)
    except Exception as e:
        raise e
    finally:
        ex.tearDown()


from pydantic import BaseModel, Field
from typing import List, Dict, Optional, Union
import asyncio
from schema_agents.role import Role, Action
from schema_agents.schema import Message
from schema_agents.teams import Team

import os
import urllib.request
import pandas as pd

class SearchQuery(BaseModel):
    """React UI Requirement
    The aim is to create a web UI for the main script in python to obtain user input, display results, and allow for interaction. 
    The exported imjoy plugin api function will be called inside the main function to interact with the user."""
    query: str = Field(..., description="Name of the React plugin for main function referencing.")
    database: str = Field(..., description="name of the databse, including positioning of elements.")


def gene_id_convert(gene_id: str) -> str:   
    """Input: database (columns), source type, source value. 
    Generate python code, execute python code, return another str (gene ensemble id type)."""
    return gene_id

def hpa_explorer(query: str) -> List[Union[str, Dict[str, str]]]:
    # check if umap_results_fit_all_transform_all.csv exists
    # if not, download from https://dl.dropbox.com/s/s4m2iysupy8gwj0/umap_results_fit_all_transform_all.csv
    # and save it to the current folder
    
    if not os.path.exists('umap_results_fit_all_transform_all.csv'):
        url = 'https://dl.dropbox.com/s/s4m2iysupy8gwj0/umap_results_fit_all_transform_all.csv'
        urllib.request.urlretrieve(url, 'umap_results_fit_all_transform_all.csv')
    df = pd.read_csv('umap_results_fit_all_transform_all.csv')

    return df.iloc[0]
 
# def pdb_explorer(query: str) -> List[Union[str, Dict[str, str]]]:
#     return ['456']


async def search(req: SearchQuery, role: Role) -> List[Union[str, Dict[str, str]]]:
    """Search for a gene in the HPA Cell Atlas and return the results."""
    
    if req.database == 'HPA Cell Atlas':
        check = await role.aask('if conversion is needed, please provide the gene id type.')
        if check:
            new_query = gene_id_convert(req.query)
        else:
            new_query = req.query
        return hpa_explorer(new_query)
    elif req.database == 'PDB':
        check = await role.aask('if conversion is needed, please provide the gene id type.')
        if check:
            new_query = gene_id_convert(req.query)
        else:
            new_query = req.query
        return pdb_explorer(new_query)
    

search_manager = Role(name="Alice",
                                profile="Search Manager",
                                goal="Search for a gene in the HPA Cell Atlas and return the results.",
                                constraints=None,
                                actions=[search])



from typing import List
from pathlib import Path


from schema_agents.const import DATA_PATH, MEM_TTL
from schema_agents.logs import logger
from schema_agents.schema import MemoryChunk
from schema_agents.utils.serialize import serialize_memory, deserialize_memory
from schema_agents.memory.faiss_store import FaissStore


class LongTermMemory(FaissStore):
    """
    The long term memory storage with Faiss as ANN search engine
    """

    def __init__(self, mem_ttl: int = MEM_TTL):
        self.role_id: str = None
        self.role_mem_path: str = None
        self.mem_ttl: int = mem_ttl  # later use
        self.threshold: float = 0.5  # experience value. TODO The threshold to filter similar memories
        self._initialized: bool = False

        self.store: 'FAISS' = None  # Faiss engine

    @property
    def is_initialized(self) -> bool:
        return self._initialized

    def recover_memory(self, role_id: str) -> List[MemoryChunk]:
        self.role_id = role_id
        self.role_mem_path = Path(DATA_PATH / f'role_mem/{self.role_id}/')
        self.role_mem_path.mkdir(parents=True, exist_ok=True)

        self.store = self._load()
        memories = []
        if not self.store:
            # TODO init `self.store` under here with raw faiss api instead under `add`
            pass
        else:
            for _id, document in self.store.docstore._dict.items():
                memories.append(deserialize_memory(document.metadata.get("memory_ser")))
            self._initialized = True

        return memories

    def _get_index_and_store_fname(self):
        if not self.role_mem_path:
            logger.error(f'You should call {self.__class__.__name__}.recover_memory fist when using LongTermMemory')
            return None, None
        index_fpath = Path(self.role_mem_path / f'{self.role_id}.index')
        storage_fpath = Path(self.role_mem_path / f'{self.role_id}.pkl')
        return index_fpath, storage_fpath

    def persist(self):
        super(LongTermMemory, self).persist()
        logger.debug(f'Agent {self.role_id} persist memory into local')

    def add(self, memory: MemoryChunk) -> bool:
        """ add memory into memory storage"""
        docs = [memory.index]
        # make sure memory.category is compatible with the BaseModel of memory.content
        # TODO
        metadatas = [{"memory_ser": serialize_memory(memory), "category": memory.category}]
        if not self.store:
            # init Faiss
            self.store = self._write(docs, metadatas)
            self._initialized = True
        else:
            self.store.add_texts(texts=docs, metadatas=metadatas)
        self.persist()
        logger.info(f"Agent {self.role_id}'s memory_storage add a piece of memory")

    def search(self, memory: MemoryChunk, k=4) -> List[MemoryChunk]:
        """search for dissimilar messages"""
        if not self.store:
            return []

        resp = self.store.similarity_search_with_score(
            query=memory.index,
            k=k
        )
        # filter the result which score is smaller than the threshold
        filtered_resp = []
        for item, score in resp:
            # the smaller score means more similar relation
            if score < self.threshold:
                continue
            # convert search result into Memory
            metadata = item.metadata
            new_mem = deserialize_memory(metadata.get("memory_ser"))
            filtered_resp.append(new_mem)
        return filtered_resp

    
    def retrieve(self, query, k=4, filter=None) -> List[MemoryChunk]:
        """retrieve relate memories from memory storage"""
        if not self.store:
            return []

        resp = self.store.similarity_search_with_score(
            query=query,
            k=k,
            filter=filter,
        )
        # filter the result which score is smaller than the threshold
        filtered_resp = []
        for item, score in resp:
            # the smaller score means more similar relation
            if score > self.threshold:
                continue
            # filter the result which category is equal to the category

            # convert search result into Memory
            metadata = item.metadata
            new_mem = deserialize_memory(metadata.get("memory_ser"))
            filtered_resp.append(new_mem)
        return filtered_resp

    def clean(self):
        index_fpath, storage_fpath = self._get_index_and_store_fname()
        if index_fpath and index_fpath.exists():
            index_fpath.unlink(missing_ok=True)
        if storage_fpath and storage_fpath.exists():
            storage_fpath.unlink(missing_ok=True)

        self.store = None
        self._initialized = False
        